# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pathlib
from io import StringIO
from typing import Optional

import numpy as np
import pandas as pd
from pandas.io.json import build_table_schema

import mlrun
import mlrun.common.schemas
import mlrun.datastore
import mlrun.utils.helpers
from mlrun.config import config as mlconf

from .base import Artifact, ArtifactSpec, StorePrefix, verify_target_path

default_preview_rows_length = 20
max_preview_columns = mlconf.artifacts.datasets.max_preview_columns
max_csv = 10000
ddf_sample_pct = 0.2
max_ddf_size = 1


class TableArtifactSpec(ArtifactSpec):
    _dict_fields = ArtifactSpec._dict_fields + ["schema", "header"]
    _exclude_fields_from_uid_hash = ArtifactSpec._exclude_fields_from_uid_hash + [
        "schema",
        "header",
    ]

    def __init__(self):
        super().__init__()
        self.schema = None
        self.header = None


class TableArtifact(Artifact):
    kind = "table"

    def __init__(
        self,
        key=None,
        body=None,
        df=None,
        viewer=None,
        visible=False,
        inline=False,
        format=None,
        header=None,
        schema=None,
    ):
        if key:
            key_suffix = pathlib.Path(key).suffix
            if not format and key_suffix:
                format = key_suffix[1:]
        super().__init__(key, body, viewer=viewer, is_inline=inline, format=format)

        if df is not None:
            self._is_df = True
            self.spec.header = df.reset_index(drop=True).columns.values.tolist()
            self.spec.format = "csv"  # todo other formats
            # if visible and not key_suffix:
            #     key += '.csv'
            self.spec._body = df
        else:
            self._is_df = False
            self.spec.header = header

        self.spec.schema = schema
        if not viewer:
            viewer = "table" if visible else None
        self.spec.viewer = viewer

    @property
    def spec(self) -> TableArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", TableArtifactSpec)

    def get_body(self):
        if not self._is_df:
            return self.spec.get_body()
        csv_buffer = StringIO()
        self.spec.get_body().to_csv(
            csv_buffer,
            encoding="utf-8",
            **mlrun.utils.line_terminator_kwargs(),
        )
        return csv_buffer.getvalue()


class DatasetArtifactSpec(ArtifactSpec):
    _dict_fields = ArtifactSpec._dict_fields + [
        "schema",
        "header",
        "length",
        "column_metadata",
        "features",
        "partition_keys",
        "timestamp_key",
        "label_column",
    ]

    _exclude_fields_from_uid_hash = ArtifactSpec._exclude_fields_from_uid_hash + [
        "schema",
        "header",
        "length",
        "column_metadata",
        "features",
        "partition_keys",
        "timestamp_key",
        "label_column",
    ]

    def __init__(self):
        super().__init__()
        self.schema = None
        self.header = None
        self.length = None
        self.column_metadata = None
        self.features = None
        self.partition_keys = None
        self.timestamp_key = None
        self.label_column = None


class DatasetArtifact(Artifact):
    kind = mlrun.common.schemas.ArtifactCategories.dataset
    # List of all the supported saving formats of a DataFrame:
    SUPPORTED_FORMATS = ["csv", "parquet", "pq", "tsdb", "kv"]
    _store_prefix = StorePrefix.Dataset

    def __init__(
        self,
        key: Optional[str] = None,
        df=None,
        preview: Optional[int] = None,
        format: str = "",  # TODO: should be changed to 'fmt'.
        stats: Optional[bool] = None,
        target_path: Optional[str] = None,
        extra_data: Optional[dict] = None,
        column_metadata: Optional[dict] = None,
        ignore_preview_limits: bool = False,
        label_column: Optional[str] = None,
        **kwargs,
    ):
        format = (format or "").lower()
        super().__init__(key, None, format=format, target_path=target_path)
        if format and format not in self.SUPPORTED_FORMATS:
            raise ValueError(
                f"Unsupported format {format} use one of {'|'.join(self.SUPPORTED_FORMATS)}"
            )

        if format == "pq":
            format = "parquet"
        self.format = format
        self.status.stats = None
        self.extra_data = extra_data or {}
        self.column_metadata = column_metadata or {}
        self.spec.label_column = label_column

        if df is not None:
            if label_column and label_column not in df.columns:
                raise mlrun.errors.MLRunValueError(
                    f"Provided dataframe doesn't include a column \"{label_column}\", so it can't be used as label"
                )
            if hasattr(df, "dask"):
                # If df is a Dask DataFrame, and it's small in-memory, convert to Pandas
                if (df.memory_usage(deep=True).sum().compute() / 1e9) < max_ddf_size:
                    df = df.compute()
            self.update_preview_fields_from_df(
                self, df, stats, preview, ignore_preview_limits
            )

        self._df = df
        self._kw = kwargs

    @property
    def spec(self) -> DatasetArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", DatasetArtifactSpec)

    def upload(self, artifact_path: Optional[str] = None):
        """
        internal, upload to target store
        :param artifact_path: required only for when generating target_path from artifact hash
        """
        if not self.spec.target_path:
            if self.spec.src_path:
                (
                    self.metadata.hash,
                    self.spec.target_path,
                ) = self.resolve_file_target_hash_path(
                    self.spec.src_path, artifact_path=artifact_path
                )
            else:
                (
                    self.metadata.hash,
                    self.spec.target_path,
                ) = self.resolve_dataframe_target_hash_path(
                    self._df, artifact_path=artifact_path
                )

        suffix = pathlib.Path(self.spec.target_path).suffix
        format = self.spec.format
        if not format:
            if suffix and suffix in [".csv", ".parquet", ".pq"]:
                format = "csv" if suffix == ".csv" else "parquet"
            else:
                format = "parquet"
        if not suffix and not self.spec.target_path.startswith("memory://"):
            self.spec.target_path = self.spec.target_path + "." + format

        if self._df is not None:
            self.spec.size, self.metadata.hash = upload_dataframe(
                self._df,
                self.spec.target_path,
                format=format,
                src_path=self.spec.src_path,
                **self._kw,
            )
        else:
            body = self.get_body()
            if body:
                self._upload_body(
                    body=body, target=self.target_path, artifact_path=artifact_path
                )
            else:
                # don't fail if no df or body
                self.spec.size, self.metadata.hash = None, None

    def resolve_dataframe_target_hash_path(self, dataframe, artifact_path: str):
        if not artifact_path:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Unable to resolve body target hash path, artifact_path is not defined"
            )
        dataframe_hash = mlrun.utils.helpers.calculate_dataframe_hash(dataframe)
        suffix = self._resolve_suffix()
        artifact_path = (
            artifact_path + "/" if not artifact_path.endswith("/") else artifact_path
        )
        target_path = f"{artifact_path}{dataframe_hash}{suffix}"
        return dataframe_hash, target_path

    @property
    def df(self) -> pd.DataFrame:
        """
        Get the dataset in this artifact.

        :return: The dataset as a DataFrame.
        """
        return self._df

    @staticmethod
    def is_format_supported(fmt: str) -> bool:
        """
        Check whether the given dataset format is supported by the DatasetArtifact.

        :param fmt: The format string to check.

        :return: True if the format is supported and False if not.
        """
        return fmt in DatasetArtifact.SUPPORTED_FORMATS

    @staticmethod
    def update_preview_fields_from_df(
        artifact, df, stats=None, preview_rows_length=None, ignore_preview_limits=False
    ):
        preview_rows_length = preview_rows_length or default_preview_rows_length
        if hasattr(df, "dask"):
            artifact.spec.length = df.shape[0].compute()
            preview_df = df.sample(frac=ddf_sample_pct).compute()
        else:
            artifact.spec.length = df.shape[0]
            preview_df = df

        if artifact.spec.length > preview_rows_length and not ignore_preview_limits:
            preview_df = df.head(preview_rows_length)

        preview_df = preview_df.reset_index()
        artifact.status.header_original_length = len(preview_df.columns)
        if len(preview_df.columns) > max_preview_columns and not ignore_preview_limits:
            preview_df = preview_df.iloc[:, :max_preview_columns]
        artifact.spec.header = preview_df.columns.values.tolist()
        artifact.status.preview = preview_df.values.tolist()
        # Table schema parsing doesn't require a column named "index"
        # to align its output with previously generated header and preview data
        if "index" in preview_df.columns:
            preview_df.drop("index", axis=1, inplace=True)
        artifact.spec.schema = build_table_schema(preview_df)

        # set artifact stats if stats is explicitly set to true, or if stats is None and the dataframe is small
        if stats or (
            stats is None
            and (
                artifact.spec.length < max_csv and len(df.columns) < max_preview_columns
            )
            or ignore_preview_limits
        ):
            artifact.status.stats = get_df_stats(df)

    @property
    def column_metadata(self):
        return self.spec.column_metadata

    @column_metadata.setter
    def column_metadata(self, column_metadata):
        self.spec.column_metadata = column_metadata

    @property
    def schema(self):
        return self.spec.schema

    @schema.setter
    def schema(self, schema):
        self.spec.schema = schema

    @property
    def header(self):
        return self.spec.header

    @header.setter
    def header(self, header):
        self.spec.header = header

    @property
    def preview(self):
        return self.status.preview

    @preview.setter
    def preview(self, preview):
        self.status.preview = preview

    @property
    def stats(self):
        return self.status.stats

    @stats.setter
    def stats(self, stats):
        self.status.stats = stats


def get_df_stats(df):
    if hasattr(df, "dask"):
        df = df.sample(frac=ddf_sample_pct).compute()
    d = {}
    for col, values in df.describe(include="all").items():
        stats_dict = {}
        for stat, val in values.dropna().items():
            if isinstance(val, (float, np.floating, np.float64)):
                stats_dict[stat] = float(val)
            elif isinstance(val, (int, np.integer, np.int64)):
                stats_dict[stat] = int(val)
            else:
                stats_dict[stat] = str(val)

        if pd.api.types.is_numeric_dtype(df[col]):
            # store histogram
            try:
                hist, bins = np.histogram(df[col], bins=20)
                stats_dict["hist"] = [hist.tolist(), bins.tolist()]
            except Exception:
                pass

        d[col] = stats_dict
    return d


def update_dataset_meta(
    artifact,
    from_df=None,
    schema: Optional[dict] = None,
    header: Optional[list] = None,
    preview: Optional[list] = None,
    stats: Optional[dict] = None,
    extra_data: Optional[dict] = None,
    column_metadata: Optional[dict] = None,
    labels: Optional[dict] = None,
    ignore_preview_limits: bool = False,
):
    """Update dataset object attributes/metadata

    this method will edit or add metadata to a dataset object

    example:
        update_dataset_meta(dataset, from_df=df,
                            extra_data={'histogram': 's3://mybucket/..'})

    :param from_df:                 read metadata (schema, preview, ..) from provided df
    :param artifact:                dataset artifact object or path (store://..) or DataItem
    :param schema:                  dataset schema, see pandas build_table_schema
    :param header:                  column headers
    :param preview:                 list of rows and row values (from df.values.tolist())
    :param stats:                   dict of column names and their stats (cleaned df.describe(include='all'))
    :param extra_data:              extra data items (key: path string | artifact)
    :param column_metadata:         dict of metadata per column
    :param labels:                  metadata labels
    :param ignore_preview_limits:   whether to ignore the preview size limits
    """

    if hasattr(artifact, "artifact_url"):
        artifact = artifact.artifact_url

    if isinstance(artifact, DatasetArtifact):
        artifact_spec = artifact
    elif mlrun.datastore.is_store_uri(artifact):
        artifact_spec, _ = mlrun.datastore.store_manager.get_store_artifact(artifact)
        verify_target_path(artifact_spec)
    else:
        raise ValueError("model path must be a model store object/URL/DataItem")

    if not artifact_spec or artifact_spec.kind != "dataset":
        raise ValueError(f"store artifact ({artifact}) is not dataset kind")

    if from_df is not None:
        DatasetArtifact.update_preview_fields_from_df(
            artifact_spec, from_df, stats, ignore_preview_limits
        )

    if header:
        artifact_spec.spec.header = header
    if stats:
        artifact_spec.status.stats = stats
    if schema:
        artifact_spec.spec.schema = schema
    if preview:
        artifact_spec.status.preview = preview
    if column_metadata:
        artifact_spec.spec.column_metadata = column_metadata
    if labels:
        for key, val in labels.items():
            artifact_spec.metadata.labels[key] = val

    if extra_data:
        artifact_spec.spec.extra_data = artifact_spec.spec.extra_data or {}
        for key, item in extra_data.items():
            if hasattr(item, "target_path"):
                item = item.spec.target_path
            artifact_spec.spec.extra_data[key] = item

    mlrun.get_run_db().store_artifact(
        artifact_spec.spec.db_key,
        artifact_spec.to_dict(),
        tree=artifact_spec.metadata.tree,
        iter=artifact_spec.metadata.iter,
        project=artifact_spec.metadata.project,
    )


def upload_dataframe(
    df, target_path, format, src_path=None, **kw
) -> tuple[Optional[int], Optional[str]]:
    if src_path and os.path.isfile(src_path):
        mlrun.datastore.store_manager.object(url=target_path).upload(src_path)
        return (
            os.stat(src_path).st_size,
            mlrun.utils.helpers.calculate_local_file_hash(src_path),
        )

    if df is None:
        return None, None

    if target_path.startswith("memory://"):
        mlrun.datastore.store_manager.object(target_path).put(df)
        return None, None

    if format in ["csv", "parquet"]:
        target_class = mlrun.datastore.targets.kind_to_driver[format]
        size = target_class(path=target_path).write_dataframe(df, **kw)
        return size, None

    raise mlrun.errors.MLRunInvalidArgumentError(f"Format {format} not implemented yet")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from urllib.parse import urlparse

import redis
import redis.cluster

import mlrun

from .base import DataStore


class RedisStore(DataStore):
    """
    Partial implementation of DataStore over the Redis KV store

    - no support for filesystem
    - key and value sizes are limited to 512MB
    """

    def __init__(
        self, parent, schema, name, endpoint="", secrets: Optional[dict] = None
    ):
        redis_default_port = "6379"
        super().__init__(parent, name, schema, endpoint, secrets=secrets)
        self.headers = None

        self.endpoint = self.endpoint or mlrun.mlconf.redis.url
        parsed_endpoint = urlparse(self.endpoint)
        if parsed_endpoint.scheme != "redis" and parsed_endpoint.scheme != "rediss":
            parsed_endpoint = urlparse(f"{schema}://{self.endpoint}")
        self.secure = parsed_endpoint.scheme == "rediss"

        if parsed_endpoint.username or parsed_endpoint.password:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Provide Redis username and password only via secrets"
            )
        user = self._get_secret_or_env("REDIS_USER", "")
        password = self._get_secret_or_env("REDIS_PASSWORD", "")
        host = parsed_endpoint.hostname
        port = parsed_endpoint.port if parsed_endpoint.port else redis_default_port
        schema = parsed_endpoint.scheme
        if user or password:
            endpoint = f"{schema}://{user}:{password}@{host}:{port}"
        else:
            endpoint = f"{schema}://{host}:{port}"
        self._redis_url = endpoint

        self._redis = None

    @property
    def redis(self):
        if self._redis is None:
            try:
                self._redis = redis.cluster.RedisCluster.from_url(
                    self._redis_url, decode_responses=True
                )
            except redis.cluster.RedisClusterException:
                self._redis = redis.Redis.from_url(
                    self._redis_url, decode_responses=True
                )

        return self._redis

    @property
    def filesystem(self):
        return None  # no support for fsspec

    def supports_isdir(self):
        return False

    @classmethod
    def build_redis_key(cls, key, prefix_only=False):
        prefixes = ["redis://", "rediss://", "ds://"]
        start = next((len(prefix) for prefix in prefixes if key.startswith(prefix)), 0)
        # skip over user/pass, host, port
        start = key.find("/", start)
        # insert the prefix '{' hashtag to the key as stored in redis
        key = "{" + key[start:]
        if prefix_only is False:
            key += "}"

        return key

    @classmethod
    def build_mlrun_key(cls, key):
        key = key[len("{") : -len("}")]

        return key

    def upload(self, key, src_path):
        key = RedisStore.build_redis_key(key)
        with open(src_path, "rb") as f:
            while True:
                data = f.read(1000 * 1000)
                if not data:
                    break
                self.redis.append(key, data)

    def get(self, key, size=None, offset=0):
        key = RedisStore.build_redis_key(key)
        if offset < 0:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "offset argument should be >= 0"
            )
        start_offset = offset
        if size is None:
            end_offset = -1
        elif size <= 0:
            raise mlrun.errors.MLRunInvalidArgumentError("size argument should be > 0")
        else:
            end_offset = start_offset + size - 1

        return self.redis.getrange(key, start_offset, end_offset)

    def put(self, key, data, append=False):
        key = RedisStore.build_redis_key(key)
        data, _ = self._prepare_put_data(data, append)
        if append:
            self.redis.append(key, data)
        else:
            self.redis.set(key, data)

    def stat(self, key):
        raise NotImplementedError()

    def listdir(self, key):
        """
        list all keys with prefix key
        """
        response = []
        key = RedisStore.build_redis_key(key, prefix_only=True)
        key += "*" if key.endswith("/") else "/*"
        for key in self.redis.scan_iter(key):
            response.append(RedisStore.build_mlrun_key(key))
        return response

    def rm(self, key, recursive=False, maxdepth=None):
        """
        delete keys, possibly recursively
        """
        if maxdepth is not None:
            raise NotImplementedError("maxdepth is not supported")

        key = RedisStore.build_redis_key(key, prefix_only=recursive)

        if recursive:
            key += "*" if key.endswith("/") else "/*"
            for k in self.redis.scan_iter(key):
                self.redis.delete(k)
            key = f"_spark:{key}"
            for k in self.redis.scan_iter(key):
                self.redis.delete(k)
        else:
            self.redis.delete(key)

    @property
    def spark_url(self):
        return ""



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Optional, Union

import mlrun

from .plan import Plan


class ArtifactsLibrary(ABC):
    """
    An abstract class for an artifacts library. Each framework should have an artifacts library for knowing what
    artifacts can be produced and their configurations. The default method must be implemented for when the user do not
    pass any plans.

    To add a plan to the library, simply write its name in the library class as a class variable pointing to the plan's
    class 'init_artifact' class method:

    some_artifact = SomeArtifactPlan
    """

    # A constant name for the context parameter to use for passing a plans configuration:
    CONTEXT_PARAMETER = "_artifacts"

    # TODO: Finish support for custom plans.
    @classmethod
    def get_plans(
        cls,
        artifacts: Optional[Union[list[Plan], dict[str, dict], list[str]]] = None,
        context: mlrun.MLClientCtx = None,
        include_default: bool = True,
        # custom_plans: dict = None, :param custom_plans: Custom user plans objects to initialize from.
        **default_kwargs,
    ) -> list[Plan]:
        """
        Get plans for a run. The plans will be taken from the provided artifacts / configuration via code, from provided
        configuration via MLRun context and if the 'include_default' is True, from the framework artifact library's
        defaults.

        :param artifacts:         The artifacts parameter passed to the function. Can be passed as a configuration
                                  dictionary or an initialized plans list that will simply be returned.
        :param context:           A context to look in if the configuration was passed as a parameter.
        :param include_default:   Whether to include the default in addition to the provided plans. Default: True.
        :param default_kwargs:    Additional key word arguments to pass to the 'default' method of the given artifact
                                  library class.

        :return: The plans list.

        :raise MLRunInvalidArgumentError: If the plans were not passed in a list or a dictionary.
        """
        # Generate the available plans dictionary:
        available_plans = cls._get_library_plans()
        # if custom_plans is not None:
        #     available_plans = {**available_plans, **custom_plans}

        # Initialize the plans list:
        parsed_plans = []  # type: List[Plan]

        # Get the user input plans:
        artifacts_from_context = None
        if context is not None:
            artifacts_from_context = context.parameters.get(cls.CONTEXT_PARAMETER, None)
        for user_input in [artifacts, artifacts_from_context]:
            if user_input is not None:
                if isinstance(user_input, dict):
                    parsed_plans += cls._from_dict(
                        requested_plans=user_input, available_plans=available_plans
                    )
                elif isinstance(user_input, list):
                    parsed_plans += cls._from_list(
                        requested_plans=user_input, available_plans=available_plans
                    )
                else:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"Artifacts plans are expected to be given in a list or a dictionary, "
                        f"got: '{type(user_input)}'."
                    )

        # Get the library's default:
        if include_default:
            parsed_plans += cls.default(**default_kwargs)

        return parsed_plans

    @classmethod
    @abstractmethod
    def default(cls, **kwargs) -> list[Plan]:
        """
        Get the default artifacts plans list of this framework's library.

        :return: The default artifacts plans list.
        """
        pass

    @classmethod
    def _get_library_plans(cls) -> dict[str, type[Plan]]:
        """
        Get all the supported plans in this library.

        :return: The library's plans.
        """
        return {  # type: Dict[str, Type[Plan]]
            plan_name: plan_class
            for plan_name, plan_class in cls.__dict__.items()
            if isinstance(plan_class, type) and not plan_name.startswith("_")
        }

    @staticmethod
    def _from_dict(
        requested_plans: dict[str, dict], available_plans: dict[str, type[Plan]]
    ) -> list[Plan]:
        """
        Initialize a list of plans from a given configuration dictionary. The configuration is expected to be a
        dictionary of plans and their initialization parameters in the following format:

        {
            PLAN_NAME: {
                PARAMETER_NAME: PARAMETER_VALUE,
            },
        }

        :param requested_plans: The configurations of plans to initialize.
        :param available_plans: The available plans to initialize from.

        :return: The initialized plans list.

        :raise MLRunInvalidArgumentError: If the configuration was incorrect due to unsupported plan or miss use of
                                          parameters in the plan initializer.
        """
        # Go through the given configuration and initialize the plans accordingly:
        plans = []  # type: List[Plan]
        for plan_name, plan_parameters in requested_plans.items():
            # Validate the plan is in the library:
            if plan_name not in available_plans:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"The given artifact '{plan_name}' is not known in this artifacts library. The known artifacts "
                    f"are: {list(available_plans.keys())}."
                )
            # Try to create the plan with the given parameters:
            try:
                plans.append(available_plans[plan_name](**plan_parameters))
            except TypeError as error:
                # A TypeError was raised, that means there was a misuse of parameters in the plan's '__init__' method:
                raise mlrun.MLRunInvalidArgumentError(
                    f"The following artifact: '{plan_name}' cannot be parsed due to misuse of parameters: {error}"
                )

        return plans

    @staticmethod
    def _from_list(
        requested_plans: list[str], available_plans: dict[str, type[Plan]]
    ) -> list[Plan]:
        """
        Initialize a list of plans from a given configuration list. The configuration is expected to be a list of plans
        names to be initialized with their default configuration.

        :param requested_plans: The plans to initialize.
        :param available_plans: The available plans to initialize from.

        :return: The initialized plans list.

        :raise MLRunInvalidArgumentError: If the configuration was incorrect due to unsupported plan.
        """
        # Go through the given configuration and initialize the plans accordingly:
        plans = []  # type: List[Plan]
        for plan in requested_plans:
            # Initialized plan:
            if isinstance(plan, Plan):
                plans.append(plan)
            # Plan name that needed to be parsed:
            elif isinstance(plan, str):
                # Validate the plan is in the library:
                if plan not in available_plans:
                    raise mlrun.errors.MLRunInvalidArgumentError(
                        f"The given artifact '{plan}' is not known in this artifacts library. The known artifacts "
                        f"are: {list(available_plans.keys())}."
                    )
                # Create the plan and collect it:
                plans.append(available_plans[plan]())
            # Unsupported type:
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"Expecting a list of artifact plans or plans names from the artifacts library but given: "
                    f"'{type(plan)}'"
                )

        return plans



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import mlrun.common.schemas
import mlrun.lists
import mlrun.utils.helpers

from .base import NotificationBase


class IPythonNotification(NotificationBase):
    """
    Client only notification for printing rich run statuses in IPython/Jupyter notebooks
    """

    def __init__(
        self,
        name: typing.Optional[str] = None,
        params: typing.Optional[dict[str, str]] = None,
        default_params: typing.Optional[dict[str, str]] = None,
    ):
        super().__init__(name, params, default_params)
        self._ipython = None
        try:
            import IPython

            if mlrun.utils.helpers.is_running_in_jupyter_notebook():
                self._ipython = IPython
        except ImportError:
            pass

    @property
    def active(self) -> bool:
        return self._ipython is not None

    def push(
        self,
        message: str,
        severity: typing.Optional[
            typing.Union[mlrun.common.schemas.NotificationSeverity, str]
        ] = mlrun.common.schemas.NotificationSeverity.INFO,
        runs: typing.Optional[typing.Union[mlrun.lists.RunList, list]] = None,
        custom_html: typing.Optional[typing.Optional[str]] = None,
        alert: typing.Optional[mlrun.common.schemas.AlertConfig] = None,
        event_data: typing.Optional[mlrun.common.schemas.Event] = None,
    ):
        if not self._ipython:
            mlrun.utils.helpers.logger.debug(
                "Not in IPython environment, skipping notification"
            )
            return

        self._ipython.display.display(
            self._ipython.display.HTML(
                self._get_html(message, severity, runs, custom_html)
            )
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import numpy as np
import onnx
import onnxruntime

import mlrun
from mlrun.serving.v2_serving import V2ModelServer

from .model_handler import ONNXModelHandler


class ONNXModelServer(V2ModelServer):
    """
    ONNX Model serving class, inheriting the V2ModelServer class for being initialized automatically by the model server
    and be able to run locally as part of a nuclio serverless function, or as part of a real-time pipeline.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        name: Optional[str] = None,
        model: onnx.ModelProto = None,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        execution_providers: Optional[
            list[Union[str, tuple[str, dict[str, Any]]]]
        ] = None,
        protocol: Optional[str] = None,
        **class_args,
    ):
        """
        Initialize a serving class for an onnx.ModelProto model.

        :param context:             For internal use (passed in init).
        :param name:                The model name to be served.
        :param model:               Model to handle or None in case a loading parameters were supplied.
        :param model_path:          Path to the model's directory to load it from. The onnx file must start with the
                                    given model name and the directory must contain the onnx file. The model path can be
                                    also passed as a model object path in the following format:
                                    'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'.
        :param model_name:          The model name for saving and logging the model:
                                    * Mandatory for loading the model from a local path.
                                    * If given a logged model (store model path) it will be read from the artifact.
                                    * If given a loaded model object and the model name is None, the name will be set to
                                      the model's object name / class.
        :param execution_providers: List of the execution providers. The first provider in the list will be the most
                                    preferred. For example, a CUDA execution provider with configurations and a CPU
                                    execution provider:
                                    [
                                        (
                                            'CUDAExecutionProvider',
                                            {
                                                'device_id': 0,
                                                'arena_extend_strategy': 'kNextPowerOfTwo',
                                                'gpu_mem_limit': 2 * 1024 * 1024 * 1024,
                                                'cudnn_conv_algo_search': 'EXHAUSTIVE',
                                                'do_copy_in_default_stream': True,
                                            }
                                        ),
                                        'CPUExecutionProvider'
                                    ]
                                    Default: None - will prefer CUDA Execution Provider over CPU Execution Provider.
        :param protocol:            -
        :param class_args:          -
        """
        super().__init__(
            context=context,
            name=name,
            model_path=model_path,
            model=model,
            protocol=protocol,
            **class_args,
        )

        # Store the model name:
        self.model_name = model_name

        # Set the execution providers (default will prefer CUDA Execution Provider over CPU Execution Provider):
        self.execution_providers = (
            ["CUDAExecutionProvider", "CPUExecutionProvider"]
            if execution_providers is None
            else execution_providers
        )

        # Prepare inference parameters:
        self._model_handler: ONNXModelHandler = None
        self._inference_session: onnxruntime.InferenceSession = None
        self._input_layers: list[str] = None
        self._output_layers: list[str] = None

    def load(self):
        """
        Use the model handler to get the model file path and initialize an ONNX run time inference session.
        """
        # Set up a model handler:
        self._model_handler = ONNXModelHandler(
            model_name=self.model_name,
            model_path=self.model_path,
            model=self.model,
            context=self.context,
        )

        # Load the model:
        if self._model_handler.model is None:
            self._model_handler.load()
        self.model = self._model_handler.model

        # initialize the onnx run time session:
        self._inference_session = onnxruntime.InferenceSession(
            onnx._serialize(self._model_handler.model),
            providers=self.execution_providers,
        )

        # Get the input layers names:
        self._input_layers = [
            input_layer.name for input_layer in self._inference_session.get_inputs()
        ]

        # Get the outputs layers names:
        self._output_layers = [
            output_layer.name for output_layer in self._inference_session.get_outputs()
        ]

    def predict(self, request: dict[str, Any]) -> np.ndarray:
        """
        Infer the inputs through the model using ONNXRunTime and return its output. The inferred data will be
        read from the "inputs" key of the request.

        :param request: The request to the model. The input to the model will be read from the "inputs" key.

        :return: The ONNXRunTime session returned output on the given inputs.
        """
        # Read the inputs from the request:
        inputs = request["inputs"]

        # Infer the inputs through the model:
        return self._inference_session.run(
            output_names=self._output_layers,
            input_feed={
                input_layer: data
                for input_layer, data in zip(self._input_layers, inputs)
            },
        )

    def explain(self, request: dict[str, Any]) -> str:
        """
        Return a string explaining what model is being serve in this serving function and the function name.

        :param request: A given request.

        :return: Explanation string.
        """
        return f"The '{self.model.name}' model serving function named '{self.name}'"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["GraphServer", "create_graph_server", "GraphContext", "MockEvent"]

import asyncio
import base64
import copy
import importlib
import json
import os
import socket
import traceback
import uuid
from collections import defaultdict
from datetime import datetime, timezone
from typing import Any, Optional, Union

import pandas as pd
import storey
from nuclio import Context as NuclioContext
from nuclio.request import Logger as NuclioLogger

import mlrun
import mlrun.common.helpers
import mlrun.common.schemas
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.datastore.datastore_profile as ds_profile
import mlrun.model_monitoring
import mlrun.utils
from mlrun.config import config
from mlrun.errors import err_to_str
from mlrun.secrets import SecretsStore

from ..common.helpers import parse_versioned_object_uri
from ..common.schemas.model_monitoring.constants import FileTargetKind
from ..common.schemas.serving import MAX_BATCH_JOB_DURATION
from ..datastore import DataItem, get_stream_pusher
from ..datastore.store_resources import ResourceCache
from ..errors import MLRunInvalidArgumentError
from ..execution import MLClientCtx
from ..model import ModelObj
from ..utils import get_caller_globals, get_relative_module_name_from_path
from .states import (
    FlowStep,
    MonitoredStep,
    RootFlowStep,
    RouterStep,
    get_function,
    graph_root_setter,
)
from .utils import event_id_key, event_path_key

DUMMY_STREAM = "dummy://"


class _StreamContext:
    """Handles the stream context for the events stream process. Includes the configuration for the output stream
    that will be used for pushing the events from the nuclio model serving function"""

    def __init__(self, enabled: bool, parameters: dict, function_uri: str):
        """
        Initialize _StreamContext object.
        :param enabled:      A boolean indication for applying the stream context
        :param parameters:   Dictionary of optional parameters, such as `log_stream` and `stream_args`. Note that these
                             parameters might be relevant to the output source such as `kafka_brokers` if
                             the output source is from type Kafka.
        :param function_uri: Full value of the function uri, usually it's <project-name>/<function-name>
        """

        self.enabled = False
        self.hostname = socket.gethostname()
        self.function_uri = function_uri
        self.output_stream = None
        log_stream = parameters.get(FileTargetKind.LOG_STREAM, "")

        if (enabled or log_stream) and function_uri:
            self.enabled = True
            project, _, _, _ = parse_versioned_object_uri(
                function_uri, config.active_project
            )

            stream_args = parameters.get("stream_args", {})

            if log_stream:
                # Get the output stream from the log stream path
                stream_path = log_stream.format(project=project)
                self.output_stream = get_stream_pusher(stream_path, **stream_args)
            else:
                # Get the output stream from the profile
                self.output_stream = mlrun.model_monitoring.helpers.get_output_stream(
                    project=project,
                    profile=parameters.get("stream_profile"),
                    mock=stream_args.get("mock", False),
                )


class GraphServer(ModelObj):
    kind = "server"

    def __init__(
        self,
        graph=None,
        parameters=None,
        load_mode=None,
        function_uri=None,
        verbose=False,
        version=None,
        functions=None,
        graph_initializer=None,
        error_stream=None,
        track_models=None,
        secret_sources=None,
        default_content_type=None,
        function_name=None,
        function_tag=None,
        project=None,
        model_endpoint_creation_task_name=None,
    ):
        self._graph = None
        self.graph: Union[RouterStep, RootFlowStep] = graph
        self.function_uri = function_uri
        self.parameters = parameters or {}
        self.verbose = verbose
        self.load_mode = load_mode or "sync"
        self.version = version or "v2"
        self.context = None
        self._current_function = None
        self.functions = functions or {}
        self.graph_initializer = graph_initializer
        self.error_stream = error_stream
        self.track_models = track_models
        self._error_stream_object = None
        self.secret_sources = secret_sources
        self._secrets = SecretsStore.from_list(secret_sources)
        self._db_conn = None
        self.resource_cache = None
        self.default_content_type = default_content_type
        self.http_trigger = True
        self.function_name = function_name
        self.function_tag = function_tag
        self.project = project
        self.model_endpoint_creation_task_name = model_endpoint_creation_task_name

    def set_current_function(self, function):
        """set which child function this server is currently running on"""
        self._current_function = function

    @property
    def graph(self) -> Union[RootFlowStep, RouterStep]:
        return self._graph

    @graph.setter
    def graph(self, graph):
        graph_root_setter(self, graph)

    def set_error_stream(self, error_stream):
        """set/initialize the error notification stream"""
        self.error_stream = error_stream
        if error_stream:
            self._error_stream_object = get_stream_pusher(error_stream)
        else:
            self._error_stream_object = None

    def _get_db(self):
        return mlrun.get_run_db(secrets=self._secrets)

    def init_states(
        self,
        context,
        namespace,
        resource_cache: Optional[ResourceCache] = None,
        logger=None,
        is_mock=False,
        monitoring_mock=False,
        stream_profile: Optional[ds_profile.DatastoreProfile] = None,
    ) -> None:
        """for internal use, initialize all steps (recursively)"""

        if self.secret_sources:
            self._secrets = SecretsStore.from_list(self.secret_sources)

        if self.error_stream:
            self._error_stream_object = get_stream_pusher(self.error_stream)
        self.resource_cache = resource_cache or ResourceCache()

        context = GraphContext(server=self, nuclio_context=context, logger=logger)
        context.is_mock = is_mock
        context.monitoring_mock = monitoring_mock
        context.root = self.graph

        if is_mock and monitoring_mock:
            if stream_profile:
                # Add the user-defined stream profile to the parameters
                self.parameters["stream_profile"] = stream_profile
            elif not (
                self.parameters.get(FileTargetKind.LOG_STREAM)
                or mlrun.get_secret_or_env(
                    mm_constants.ProjectSecretKeys.STREAM_PROFILE_NAME
                )
            ):
                # Set a dummy log stream for mocking purposes if there is no direct
                # user-defined stream profile and no information in the environment
                self.parameters[FileTargetKind.LOG_STREAM] = DUMMY_STREAM

        context.stream = _StreamContext(
            self.track_models, self.parameters, self.function_uri
        )
        context.current_function = self._current_function
        context.get_store_resource = self.resource_cache.resource_getter(
            self._get_db(), self._secrets
        )
        context.get_table = self.resource_cache.get_table
        context.verbose = self.verbose
        self.context = context

        if self.graph_initializer:
            if callable(self.graph_initializer):
                handler = self.graph_initializer
            else:
                handler = get_function(self.graph_initializer, namespace or [])
            handler(self)

        context.root = self.graph

    def init_object(self, namespace):
        self.graph.init_object(self.context, namespace, self.load_mode, reset=True)

    def test(
        self,
        path: str = "/",
        body: Optional[Union[str, bytes, dict]] = None,
        method: str = "",
        headers: Optional[str] = None,
        content_type: Optional[str] = None,
        silent: bool = False,
        get_body: bool = True,
        event_id: Optional[str] = None,
        trigger: "MockTrigger" = None,
        offset=None,
        time=None,
    ):
        """invoke a test event into the server to simulate/test server behavior

        example::

            server = create_graph_server()
            server.add_model("my", class_name=MyModelClass, model_path="{path}", z=100)
            print(server.test("my/infer", testdata))

        :param path:       api path, e.g. (/{router.url_prefix}/{model-name}/..) path
        :param body:       message body (dict or json str/bytes)
        :param method:     optional, GET, POST, ..
        :param headers:    optional, request headers, ..
        :param content_type:  optional, http mime type
        :param silent:     don't raise on error responses (when not 20X)
        :param get_body:   return the body as py object (vs serialize response into json)
        :param event_id:   specify the unique event ID (by default a random value will be generated)
        :param trigger:    nuclio trigger info or mlrun.serving.server.MockTrigger class (holds kind and name)
        :param offset:     trigger offset (for streams)
        :param time:       event time Datetime or str, default to now()
        """
        if not self.graph:
            raise MLRunInvalidArgumentError(
                "no models or steps were set, use function.set_topology() and add steps"
            )
        if not method:
            method = "POST" if body else "GET"
        event = MockEvent(
            body=body,
            path=path,
            method=method,
            headers=headers,
            content_type=content_type,
            event_id=event_id,
            trigger=trigger,
            offset=offset,
            time=time,
        )
        resp = self.run(event, get_body=get_body)
        if hasattr(resp, "status_code") and resp.status_code >= 300 and not silent:
            raise RuntimeError(f"failed ({resp.status_code}): {resp.body}")
        return resp

    def run(self, event, context=None, get_body=False, extra_args=None):
        server_context = self.context
        context = context or server_context
        event.content_type = event.content_type or self.default_content_type or ""
        if event.headers:
            if event_id_key in event.headers:
                event.id = event.headers.get(event_id_key)
            if event_path_key in event.headers:
                event.path = event.headers.get(event_path_key)

        if isinstance(event.body, (str, bytes)) and (
            not event.content_type or event.content_type in ["json", "application/json"]
        ):
            # assume it is json and try to load
            try:
                body = json.loads(event.body)
                event.body = body
            except (json.decoder.JSONDecodeError, UnicodeDecodeError) as exc:
                if event.content_type in ["json", "application/json"]:
                    # if its json type and didnt load, raise exception
                    message = f"failed to json decode event, {err_to_str(exc)}"
                    context.logger.error(message)
                    server_context.push_error(event, message, source="_handler")
                    return context.Response(
                        body=message, content_type="text/plain", status_code=400
                    )
        try:
            response = self.graph.run(event, **(extra_args or {}))
        except Exception as exc:
            message = f"{exc.__class__.__name__}: {err_to_str(exc)}"
            if server_context.verbose:
                message += "\n" + str(traceback.format_exc())
            context.logger.error(f"run error, {traceback.format_exc()}")
            server_context.push_error(event, message, source="_handler")
            return context.Response(
                body=message, content_type="text/plain", status_code=400
            )

        if asyncio.iscoroutine(response):
            return self._process_async_response(context, response, get_body)
        else:
            return self._process_response(context, response, get_body)

    async def _process_async_response(self, context, response, get_body):
        return self._process_response(context, await response, get_body)

    def _process_response(self, context, response, get_body):
        body = response.body
        if (
            isinstance(context, MLClientCtx)
            or isinstance(body, context.Response)
            or get_body
        ):
            return body

        if body and not isinstance(body, (str, bytes)):
            body = json.dumps(body)
            return context.Response(
                body=body, content_type="application/json", status_code=200
            )
        return body

    def wait_for_completion(self):
        """wait for async operation to complete"""
        return self.graph.wait_for_completion()


def add_error_raiser_step(
    graph: RootFlowStep, monitored_steps: dict[str, MonitoredStep]
) -> RootFlowStep:
    monitored_steps_raisers = {}
    user_steps = list(graph.steps.values())
    for monitored_step in monitored_steps.values():
        error_step = graph.add_step(
            class_name="mlrun.serving.states.ModelRunnerErrorRaiser",
            name=f"{monitored_step.name}_error_raise",
            after=monitored_step.name,
            full_event=True,
            raise_exception=monitored_step.raise_exception,
            models_names=list(monitored_step.class_args["models"].keys()),
            model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
            function=monitored_step.function,
        )
        if monitored_step.responder:
            monitored_step.responder = False
            error_step.respond()
        monitored_steps_raisers[monitored_step.name] = error_step.name
        error_step.on_error = monitored_step.on_error
    if monitored_steps_raisers:
        for step in user_steps:
            if step.after:
                if isinstance(step.after, list):
                    for i in range(len(step.after)):
                        if step.after[i] in monitored_steps_raisers:
                            step.after[i] = monitored_steps_raisers[step.after[i]]
                else:
                    if (
                        isinstance(step.after, str)
                        and step.after in monitored_steps_raisers
                    ):
                        step.after = monitored_steps_raisers[step.after]
    return graph


def add_monitoring_general_steps(
    project: str,
    graph: RootFlowStep,
    context,
    serving_spec,
    pause_until_background_task_completion: bool,
) -> tuple[RootFlowStep, FlowStep]:
    """
    Adding the monitoring flow connection steps, this steps allow the graph to reconstruct the serving event enrich it
    and push it to the model monitoring stream
    system_steps structure -
        "background_task_status_step" --> "filter_none" --> "monitoring_pre_processor_step" --> "flatten_events"
        --> "sampling_step" --> "filter_none_sampling" --> "model_monitoring_stream"
    """
    background_task_status_step = None
    if pause_until_background_task_completion:
        background_task_status_step = graph.add_step(
            "mlrun.serving.system_steps.BackgroundTaskStatus",
            "background_task_status_step",
            model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
            full_event=True,
        )
    monitor_flow_step = graph.add_step(
        "storey.Filter",
        "filter_none",
        _fn="(event is not None)",
        after="background_task_status_step" if background_task_status_step else None,
        model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
    )
    if background_task_status_step:
        monitor_flow_step = background_task_status_step
    graph.add_step(
        "mlrun.serving.system_steps.MonitoringPreProcessor",
        "monitoring_pre_processor_step",
        after="filter_none",
        full_event=True,
        model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
    )
    # flatten the events
    graph.add_step(
        "storey.FlatMap",
        "flatten_events",
        _fn="(event)",
        after="monitoring_pre_processor_step",
        model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
    )
    graph.add_step(
        "mlrun.serving.system_steps.SamplingStep",
        "sampling_step",
        after="flatten_events",
        sampling_percentage=float(
            serving_spec.get("parameters", {}).get("sampling_percentage", 100.0)
            if isinstance(serving_spec, dict)
            else getattr(serving_spec, "parameters", {}).get(
                "sampling_percentage", 100.0
            ),
        ),
        model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
    )
    graph.add_step(
        "storey.Filter",
        "filter_none_sampling",
        _fn="(event is not None)",
        after="sampling_step",
        model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
    )

    if getattr(context, "is_mock", False):
        graph.add_step(
            "mlrun.serving.system_steps.MockStreamPusher",
            "model_monitoring_stream",
            after="filter_none_sampling",
            model_endpoint_creation_strategy=mlrun.common.schemas.ModelEndpointCreationStrategy.SKIP,
        )
    else:
        stream_uri = mlrun.model_monitoring.get_stream_path(
            project=project,
            function_name=mlrun.common.schemas.MonitoringFunctionNames.STREAM,
        )
        context.logger.info_with(
            "Creating Model Monitoring stream target using uri:", uri=stream_uri
        )
        graph.add_step(
            ">>",
            "model_monitoring_stream",
            path=stream_uri,
            sharding_func=mlrun.common.schemas.model_monitoring.constants.StreamProcessingEvent.ENDPOINT_ID,
            after="filter_none_sampling",
        )
    return graph, monitor_flow_step


def add_system_steps_to_graph(
    project: str,
    graph: RootFlowStep,
    track_models: bool,
    context,
    serving_spec,
    pause_until_background_task_completion: bool = True,
) -> RootFlowStep:
    if not (isinstance(graph, RootFlowStep) and graph.include_monitored_step()):
        return graph
    monitored_steps = graph.get_monitored_steps()
    graph = add_error_raiser_step(graph, monitored_steps)
    if track_models:
        background_task_status_step = None
        graph, monitor_flow_step = add_monitoring_general_steps(
            project,
            graph,
            context,
            serving_spec,
            pause_until_background_task_completion,
        )
        if background_task_status_step:
            monitor_flow_step = background_task_status_step
        # Connect each model runner to the monitoring step:
        for step_name, step in monitored_steps.items():
            if monitor_flow_step.after:
                if isinstance(monitor_flow_step.after, list):
                    monitor_flow_step.after.append(step_name)
                elif isinstance(monitor_flow_step.after, str):
                    monitor_flow_step.after = [monitor_flow_step.after, step_name]
            else:
                monitor_flow_step.after = [
                    step_name,
                ]
    return graph


def v2_serving_init(context, namespace=None):
    """hook for nuclio init_context()"""

    context.logger.info("Initializing server from spec")
    spec = mlrun.utils.get_serving_spec()
    server = GraphServer.from_dict(spec)
    server.graph = add_system_steps_to_graph(
        server.project,
        copy.deepcopy(server.graph),
        spec.get("track_models"),
        context,
        spec,
    )

    if config.log_level.lower() == "debug":
        server.verbose = True
    if hasattr(context, "trigger"):
        server.http_trigger = getattr(context.trigger, "kind", "http") == "http"
    context.logger.info_with(
        "Setting current function",
        current_function=os.getenv("SERVING_CURRENT_FUNCTION", ""),
    )
    server.set_current_function(os.getenv("SERVING_CURRENT_FUNCTION", ""))
    context.logger.info_with(
        "Initializing states", namespace=namespace or get_caller_globals()
    )
    kwargs = {}
    if hasattr(context, "is_mock"):
        kwargs["is_mock"] = context.is_mock
    server.init_states(
        context,
        namespace or get_caller_globals(),
        **kwargs,
    )
    context.logger.info("Initializing graph steps")
    server.init_object(namespace or get_caller_globals())
    # set the handler hook to point to our handler
    setattr(context, "mlrun_handler", v2_serving_handler)
    setattr(context, "_server", server)
    context.logger.info_with("Serving was initialized", verbose=server.verbose)
    if server.verbose:
        context.logger.info(server.to_yaml())

    _set_callbacks(server, context)


async def async_execute_graph(
    context: MLClientCtx,
    data: DataItem,
    timestamp_column: Optional[str],
    batching: bool,
    batch_size: Optional[int],
    read_as_lists: bool,
    nest_under_inputs: bool,
) -> None:
    # Validate that data parameter is a DataItem and not passed via params
    if not isinstance(data, DataItem):
        raise MLRunInvalidArgumentError(
            f"Parameter 'data' has type hint 'DataItem' but got {type(data).__name__} instead. "
            f"Data files and artifacts must be passed via the 'inputs' parameter, not 'params'. "
            f"The 'params' parameter is for simple configuration values (strings, numbers, booleans), "
            f"while 'inputs' is for data files that need to be loaded. "
            f"Example: run_function(..., inputs={{'data': 'path/to/data.csv'}}, params={{other_config: value}})"
        )
    run_call_count = 0
    spec = mlrun.utils.get_serving_spec()
    modname = None
    code = os.getenv("MLRUN_EXEC_CODE")
    if code:
        code = base64.b64decode(code).decode("utf-8")
        with open("user_code.py", "w") as fp:
            fp.write(code)
        modname = "user_code"
    else:
        # TODO: find another way to get the local file path, or ensure that MLRUN_EXEC_CODE
        #  gets set in local flow and not just in the remote pod
        source_file_path = spec.get("filename", None)
        if source_file_path:
            source_file_path_object, working_dir_path_object = (
                mlrun.utils.helpers.get_source_and_working_dir_paths(source_file_path)
            )
            if not source_file_path_object.is_relative_to(working_dir_path_object):
                raise mlrun.errors.MLRunRuntimeError(
                    f"Source file path '{source_file_path}' is not under the current working directory "
                    f"(which is required when running with local=True)"
                )
            modname = get_relative_module_name_from_path(
                source_file_path_object, working_dir_path_object
            )

    namespace = {}
    if modname:
        mod = importlib.import_module(modname)
        namespace = mod.__dict__

    server = GraphServer.from_dict(spec)

    if server.model_endpoint_creation_task_name:
        context.logger.info(
            f"Waiting for model endpoint creation task '{server.model_endpoint_creation_task_name}'..."
        )
        background_task = (
            mlrun.get_run_db().wait_for_background_task_to_reach_terminal_state(
                project=server.project,
                name=server.model_endpoint_creation_task_name,
            )
        )
        task_state = background_task.status.state
        if task_state == mlrun.common.schemas.BackgroundTaskState.failed:
            raise mlrun.errors.MLRunRuntimeError(
                "Aborting job due to model endpoint creation background task failure"
            )
        elif task_state != mlrun.common.schemas.BackgroundTaskState.succeeded:
            # this shouldn't happen, but we need to know if it does
            raise mlrun.errors.MLRunRuntimeError(
                "Aborting job because the model endpoint creation background task did not succeed "
                f"(status='{task_state}')"
            )

    df = data.as_df()

    if df.empty:
        context.logger.warn("Job terminated due to empty inputs (0 rows)")
        return []

    track_models = spec.get("track_models")

    if track_models and timestamp_column:
        context.logger.info(f"Sorting dataframe by {timestamp_column}")
        df[timestamp_column] = pd.to_datetime(  # in case it's a string
            df[timestamp_column]
        )
        df.sort_values(by=timestamp_column, inplace=True)
        if len(df) > 1:
            start_time = df[timestamp_column].iloc[0]
            end_time = df[timestamp_column].iloc[-1]
            time_range = end_time - start_time
            start_time = start_time.isoformat()
            end_time = end_time.isoformat()
            # TODO: tie this to the controller's base period
            if time_range > pd.Timedelta(MAX_BATCH_JOB_DURATION):
                raise mlrun.errors.MLRunRuntimeError(
                    f"Dataframe time range is too long: {time_range}. "
                    "Please disable tracking or reduce the input dataset's time range below the defined limit "
                    f"of {MAX_BATCH_JOB_DURATION}."
                )
        else:
            start_time = end_time = df["timestamp"].iloc[0].isoformat()
    else:
        # end time will be set from clock time when the batch completes
        start_time = datetime.now(tz=timezone.utc).isoformat()

    server.graph = add_system_steps_to_graph(
        server.project,
        copy.deepcopy(server.graph),
        track_models,
        context,
        spec,
        pause_until_background_task_completion=False,  # we've already awaited it
    )

    if config.log_level.lower() == "debug":
        server.verbose = True
    kwargs = {}
    if hasattr(context, "is_mock"):
        kwargs["is_mock"] = context.is_mock
    server.init_states(
        context=None,  # this context is expected to be a nuclio context, which we don't have in this flow
        namespace=namespace,
        **kwargs,
    )
    context.logger.info("Initializing graph steps")
    server.init_object(namespace)

    context.logger.info_with("Graph was initialized", verbose=server.verbose)

    if server.verbose:
        context.logger.info(server.to_yaml())

    async def run(body):
        nonlocal run_call_count
        event = storey.Event(id=index, body=body)
        if timestamp_column:
            if batching:
                # we use the first row in the batch to determine the timestamp for the whole batch
                body = body[0]
            if not isinstance(body, dict):
                raise mlrun.errors.MLRunRuntimeError(
                    f"When timestamp_column=True, event body must be a dict – got {type(body).__name__} instead"
                )
            if timestamp_column not in body:
                raise mlrun.errors.MLRunRuntimeError(
                    f"Event body '{body}' did not contain timestamp column '{timestamp_column}'"
                )
            event._original_timestamp = body[timestamp_column]
        run_call_count += 1
        return await server.run(event, context)

    if batching and not batch_size:
        batch_size = len(df)

    batch = []
    tasks = []
    for index, row in df.iterrows():
        data = row.to_list() if read_as_lists else row.to_dict()
        if nest_under_inputs:
            data = {"inputs": data}
        if batching:
            batch.append(data)
            if len(batch) == batch_size:
                tasks.append(asyncio.create_task(run(batch)))
                batch = []
        else:
            tasks.append(asyncio.create_task(run(data)))

    if batch:
        tasks.append(asyncio.create_task(run(batch)))

    responses = await asyncio.gather(*tasks)

    termination_result = server.wait_for_completion()
    if asyncio.iscoroutine(termination_result):
        await termination_result

    model_endpoint_uids = spec.get("model_endpoint_uids", [])

    # needed for output_stream to be created
    server = GraphServer.from_dict(spec)
    server.init_states(None, namespace)

    batch_completion_time = datetime.now(tz=timezone.utc).isoformat()

    if not timestamp_column:
        end_time = batch_completion_time

    mm_stream_record = dict(
        kind="batch_complete",
        project=context.project,
        first_timestamp=start_time,
        last_timestamp=end_time,
        batch_completion_time=batch_completion_time,
    )
    output_stream = server.context.stream.output_stream
    for mep_uid in spec.get("model_endpoint_uids", []):
        mm_stream_record["endpoint_id"] = mep_uid
        output_stream.push(mm_stream_record, partition_key=mep_uid)

    context.logger.info(
        f"Job completed processing {len(df)} rows",
        timestamp_column=timestamp_column,
        model_endpoint_uids=model_endpoint_uids,
    )

    # log the results as artifacts
    num_of_meps_in_the_graph = len(server.graph.model_endpoints_names)
    artifact_path = None
    if (
        "{{run.uid}}" not in context.artifact_path
    ):  # TODO: delete when IG-22841 is resolved
        artifact_path = "+/{{run.uid}}"  # will be concatenated to the context's path in extend_artifact_path
    if num_of_meps_in_the_graph <= 1:
        context.log_dataset(
            "prediction", df=pd.DataFrame(responses), artifact_path=artifact_path
        )
    else:
        # turn this list of samples into a dict of lists, one per model endpoint
        grouped = defaultdict(list)
        for sample in responses:
            for model_name, features in sample.items():
                grouped[model_name].append(features)
        # create a dataframe per model endpoint and log it
        for model_name, features in grouped.items():
            context.log_dataset(
                f"prediction_{model_name}",
                df=pd.DataFrame(features),
                artifact_path=artifact_path,
            )
    context.log_result("num_rows", run_call_count)


def _is_inside_asyncio_loop():
    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False


# Workaround for running with local=True in Jupyter (ML-10620)
def _workaround_asyncio_nesting():
    try:
        import nest_asyncio
    except ImportError:
        raise mlrun.errors.MLRunRuntimeError(
            "Cannot execute graph from within an already running asyncio loop. "
            "Attempt to import nest_asyncio as a workaround failed as well."
        )
    nest_asyncio.apply()


def execute_graph(
    context: MLClientCtx,
    data: DataItem,
    timestamp_column: Optional[str] = None,
    batching: bool = False,
    batch_size: Optional[int] = None,
    read_as_lists: bool = False,
    nest_under_inputs: bool = False,
) -> (list[Any], Any):
    """
    Execute graph as a job, from start to finish.

    :param context: The job's execution client context.
    :param data: The input data to the job, to be pushed into the graph row by row, or in batches.
    :param timestamp_column: The name of the column that will be used as the timestamp for model monitoring purposes.
        when timestamp_column is used in conjunction with batching, the first timestamp will be used for the entire
        batch.
    :param batching: Whether to push one or more batches into the graph rather than row by row.
    :param batch_size: The number of rows to push per batch. If not set, and batching=True, the entire dataset will
        be pushed into the graph in one batch.
    :param read_as_lists: Whether to read each row as a list instead of a dictionary.
    :param nest_under_inputs: Whether to wrap each row with {"inputs": ...}.

    :return: A list of responses.
    """
    if _is_inside_asyncio_loop():
        _workaround_asyncio_nesting()

    return asyncio.run(
        async_execute_graph(
            context,
            data,
            timestamp_column,
            batching,
            batch_size,
            read_as_lists,
            nest_under_inputs,
        )
    )


def _set_callbacks(server, context):
    if not server.graph.supports_termination() or not hasattr(context, "platform"):
        return

    if hasattr(context.platform, "set_termination_callback"):
        context.logger.info(
            "Setting termination callback to terminate graph on worker shutdown"
        )

        async def termination_callback():
            context.logger.info("Termination callback called")
            maybe_coroutine = server.wait_for_completion()
            if asyncio.iscoroutine(maybe_coroutine):
                await maybe_coroutine
            context.logger.info("Termination of async flow is completed")

        context.platform.set_termination_callback(termination_callback)

    if hasattr(context.platform, "set_drain_callback"):
        context.logger.info(
            "Setting drain callback to terminate and restart the graph on a drain event (such as rebalancing)"
        )

        async def drain_callback():
            context.logger.info("Drain callback called")
            maybe_coroutine = server.wait_for_completion()
            if asyncio.iscoroutine(maybe_coroutine):
                await maybe_coroutine
            context.logger.info(
                "Termination of async flow is completed. Rerunning async flow."
            )
            # Rerun the flow without reconstructing it
            server.graph._run_async_flow()
            context.logger.info("Async flow restarted")

        context.platform.set_drain_callback(drain_callback)


def v2_serving_handler(context, event, get_body=False):
    """hook for nuclio handler()"""
    if context._server.http_trigger:
        # Workaround for a Nuclio bug where it sometimes passes b'' instead of None due to dirty memory
        if event.body == b"":
            event.body = None

    # original path is saved in stream_path so it can be used by explicit ack, but path is reset to / as a
    # workaround for NUC-178
    # nuclio 1.12.12 added the topic attribute, and we must use it as part of the fix for NUC-233
    # TODO: Remove fallback on event.path once support for nuclio<1.12.12 is dropped
    event.stream_path = getattr(event, "topic", event.path)
    if hasattr(event, "trigger") and event.trigger.kind in (
        "kafka",
        "kafka-cluster",
        "v3ioStream",
        "v3io-stream",
        "rabbit-mq",
        "rabbitMq",
    ):
        event.path = "/"

    return context._server.run(event, context, get_body)


def create_graph_server(
    parameters=None,
    load_mode=None,
    graph=None,
    verbose=False,
    current_function=None,
    **kwargs,
) -> GraphServer:
    """create graph server host/emulator for local or test runs

    Usage example::

        server = create_graph_server(graph=RouterStep(), parameters={})
        server.init(None, globals())
        server.graph.add_route("my", class_name=MyModelClass, model_path="{path}", z=100)
        print(server.test("/v2/models/my/infer", testdata))
    """
    parameters = parameters or {}
    server = GraphServer(graph, parameters, load_mode, verbose=verbose, **kwargs)
    server.set_current_function(
        current_function or os.getenv("SERVING_CURRENT_FUNCTION", "")
    )
    return server


class MockTrigger:
    """mock nuclio event trigger"""

    def __init__(self, kind="", name=""):
        self.kind = kind
        self.name = name


class MockEvent:
    """mock basic nuclio event object"""

    def __init__(
        self,
        body=None,
        content_type=None,
        headers=None,
        method=None,
        path=None,
        event_id=None,
        trigger: MockTrigger = None,
        offset=None,
        time=None,
    ):
        self.id = event_id or uuid.uuid4().hex
        self.key = ""
        self.body = body

        # optional
        self.headers = headers or {}
        self.method = method
        self.path = path or "/"
        self.content_type = content_type
        self.error = None
        self.trigger = trigger or MockTrigger()
        self.offset = offset or 0

    def __str__(self):
        error = f", error={self.error}" if self.error else ""
        return f"Event(id={self.id}, body={self.body}, method={self.method}, path={self.path}{error})"


class Response:
    def __init__(self, headers=None, body=None, content_type=None, status_code=200):
        self.headers = headers or {}
        self.body = body
        self.status_code = status_code
        self.content_type = content_type or "text/plain"

    def __repr__(self):
        cls = self.__class__.__name__
        items = self.__dict__.items()
        args = (f"{key}={repr(value)}" for key, value in items)
        args_str = ", ".join(args)
        return f"{cls}({args_str})"


class GraphContext:
    """Graph context object"""

    def __init__(
        self,
        level="info",  # Unused argument
        logger=None,
        server=None,
        nuclio_context: Optional[NuclioContext] = None,
    ) -> None:
        self.state = None
        self.logger = logger
        self.worker_id = 0
        self.Response = Response
        self.verbose = False
        self.stream = None
        self.root = None
        self.executor: Optional[storey.flow.RunnableExecutor] = None

        if nuclio_context:
            self.logger: NuclioLogger = nuclio_context.logger
            self.Response = nuclio_context.Response
            if hasattr(nuclio_context, "trigger") and hasattr(
                nuclio_context.trigger, "kind"
            ):
                self.trigger = nuclio_context.trigger.kind
            self.worker_id = nuclio_context.worker_id
            if hasattr(nuclio_context, "platform"):
                self.platform = nuclio_context.platform
        elif not logger:
            self.logger: mlrun.utils.Logger = mlrun.utils.logger

        self._server = server
        self.current_function = None
        self.get_store_resource = None
        self.get_table = None
        self.is_mock = False
        self.monitoring_mock = False
        self._project_obj = None

    @property
    def server(self):
        return self._server

    @property
    def project_obj(self):
        if not self._project_obj:
            self._project_obj = mlrun.get_run_db().get_project(name=self.project)
        return self._project_obj

    @property
    def project(self) -> str:
        """current project name (for the current function)"""
        project, _, _, _ = mlrun.common.helpers.parse_versioned_object_uri(
            self._server.function_uri
        )
        return project

    def push_error(self, event, message, source=None, **kwargs):
        if self.verbose:
            self.logger.error(
                f"got error from {source} state:\n{event.body}\n{message}"
            )
        if self._server and self._server._error_stream_object:
            try:
                message = format_error(
                    self._server, self, source, event, message, kwargs
                )
                self._server._error_stream_object.push(message)
            except Exception as ex:
                message = traceback.format_exc()
                self.logger.error(f"failed to write to error stream: {ex}\n{message}")

    def get_param(self, key: str, default=None):
        if self._server and self._server.parameters:
            return self._server.parameters.get(key, default)
        return default

    def get_secret(self, key: str):
        if self._server and self._server._secrets:
            return self._server._secrets.get(key)
        return None

    def get_remote_endpoint(self, name, external=True):
        """return the remote nuclio/serving function http(s) endpoint given its name

        :param name: the function name/uri in the form [project/]function-name[:tag]
        :param external: return the external url (returns the external url by default)
        """
        if "://" in name:
            return name
        project, uri, tag, _ = mlrun.common.helpers.parse_versioned_object_uri(
            self._server.function_uri
        )
        if name.startswith("."):
            name = f"{uri}-{name[1:]}"
        else:
            project, name, tag, _ = mlrun.common.helpers.parse_versioned_object_uri(
                name, project
            )
        (
            state,
            fullname,
            _,
            _,
            _,
            function_status,
        ) = mlrun.runtimes.nuclio.function.get_nuclio_deploy_status(name, project, tag)

        if state in ["error", "unhealthy"]:
            raise ValueError(
                f"Nuclio function {fullname} is in error state, cannot be accessed"
            )

        key = "externalInvocationUrls" if external else "internalInvocationUrls"
        urls = function_status.get(key)
        if not urls:
            raise ValueError(f"cannot read {key} for nuclio function {fullname}")
        return f"http://{urls[0]}"


def format_error(server, context, source, event, message, args):
    return {
        "function_uri": server.function_uri,
        "worker": context.worker_id,
        "host": socket.gethostname(),
        "source": source,
        "event": {"id": event.id, "body": event.body},
        "message": message,
        "args": args,
    }



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "V2ModelServer",
    "VotingEnsemble",
    "GraphServer",
    "create_graph_server",
    "GraphContext",
    "TaskStep",
    "RouterStep",
    "QueueStep",
    "ErrorStep",
    "MonitoringApplicationStep",
    "ModelRunnerStep",
    "ModelRunner",
    "Model",
    "ModelSelector",
    "MonitoredStep",
    "LLModel",
]

from .routers import ModelRouter, VotingEnsemble  # noqa
from .server import (
    GraphContext,
    GraphServer,
    create_graph_server,
)  # noqa
from .states import (
    ErrorStep,
    QueueStep,
    RouterStep,
    TaskStep,
    MonitoringApplicationStep,
    ModelRunnerStep,
    ModelRunner,
    Model,
    ModelSelector,
    MonitoredStep,
    LLModel,
)  # noqa
from .v1_serving import MLModelServer, new_v1_model_server  # noqa
from .v2_serving import V2ModelServer  # noqa



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from abc import ABC, abstractmethod
from enum import Enum

from IPython.display import HTML, display

import mlrun

from .._common import Plan
from .utils import MLTypes


class MLPlanStages(Enum):
    """
    Stages for a machine learning plan to be produced.
    """

    # SciKit-Learn's API:
    PRE_FIT = "pre_fit"
    POST_FIT = "post_fit"
    PRE_PREDICT = "pre_predict"
    POST_PREDICT = "post_predict"

    # Boosting API:
    PRE_TRAIN = "pre_train"
    POST_TRAIN = "post_train"
    PRE_ITERATION = "pre_iteration"
    POST_ITERATION = "post_iteration"


class MLPlan(Plan, ABC):
    """
    An abstract class for describing a ML plan. A ML plan is used to produce artifact manually or in a given time using
    the MLLogger.
    """

    def __init__(self, need_probabilities: bool = False):
        """
        Initialize a new ML plan.

        :param need_probabilities: Whether this plan will need the predictions return from 'model.predict()' or
                                   'model.predict_proba()'. True means predict_proba and False predict. Default:
                                   False.
        """
        self._need_probabilities = need_probabilities
        super().__init__()

    @property
    def need_probabilities(self) -> bool:
        """
        Whether this plan require predictions returned from 'model.predict()' or 'model.predict_proba()'.

        :return: True if predict_proba and False if predict.
        """
        return self._need_probabilities

    @abstractmethod
    def is_ready(self, stage: MLPlanStages, is_probabilities: bool) -> bool:
        """
        Check whether or not the plan is fit for production by the given stage and prediction probabilities.

        :param stage:            The stage to check if the plan is ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not.

        :return: True if the plan is producible and False otherwise.
        """
        pass


class MLPlotPlan(MLPlan, ABC):
    """
    An abstract class for describing a ML plan for plots. A ML plan is used to produce artifact manually or in a given
    time using the MLLogger.
    """

    def _cli_display(self):
        """
        Print the logged artifacts names and their URIs.
        """
        print(
            json.dumps(
                {name: artifact.uri for name, artifact in self._artifacts.items()}
            )
        )

    def _gui_display(self):
        """
        Plot the plot.
        """
        for artifact in self._artifacts.values():
            if artifact.kind == "plotly":
                display(HTML(artifact.get_body()))

    def _calculate_predictions(
        self,
        y_pred: MLTypes.DatasetType = None,
        model: MLTypes.ModelType = None,
        x: MLTypes.DatasetType = None,
    ) -> MLTypes.DatasetType:
        """
        Calculate the predictions using the model and input dataset only if the predictions (y_pred) were not provided.

        :param y_pred: Predictions if available to not run predict / predict_proba.
        :param model:  The model to predict with.
        :param x:      The input dataset to predict on.

        :return: The predictions.

        :raise MLRunInvalidArgumentError: If model or x were not provided to calculate the predictions.
        """
        if y_pred is None:
            if model is None or x is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "In order to produce a confusion matrix artifact, the predictions must be passed. If predictions "
                    "are not available, the model and a dataset (x) must be given to produce them."
                )
            if self._need_probabilities:
                y_pred = (
                    model.original_predict_proba(x)
                    if hasattr(model, "original_predict_proba")
                    else model.predict_proba(x)
                )
            else:
                y_pred = (
                    model.original_predict(x)
                    if hasattr(model, "original_predict")
                    else model.predict(x)
                )

        return y_pred



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
import os
import shutil
import traceback
from typing import Any, Optional, Union

import mlrun.errors
from mlrun.artifacts import Artifact
from mlrun.artifacts.base import verify_target_path
from mlrun.datastore import DataItem, get_store_resource, store_manager
from mlrun.errors import MLRunInvalidArgumentError
from mlrun.utils import logger

from .errors import (
    MLRunPackageCollectionError,
    MLRunPackagePackingError,
    MLRunPackageUnpackingError,
)
from .packager import Packager
from .packagers.default_packager import DefaultPackager
from .utils import LogHintKey, TypeHintUtils


class PackagersManager:
    """
    A packager manager holds the project's packagers and sends them objects to pack, and data items to unpack.

    It prepares the instructions / log hint configurations and then looks for the first packager that fits the task.
    """

    def __init__(self, default_packager: Optional[type[Packager]] = None):
        """
        Initialize a packagers manager.

        :param default_packager: The default packager should be a packager that fits all types. It
                                 should fit any packagers that are managed by the manager that do not fit an
                                 object or data item. Default to ``mlrun.DefaultPackager``.
        """
        # Set the default packager:
        self._default_packager = (default_packager or DefaultPackager)()

        # Initialize the packagers list (with the default packager in it):
        self._packagers: list[Packager] = []

        # Set an artifacts list and results dictionary to collect all packed objects (will be used later to write extra
        # data if noted by the user using the log hint key "extra_data")
        self._artifacts: list[Artifact] = []
        self._results = {}

    @property
    def artifacts(self) -> list[Artifact]:
        """
        Get the artifacts that were packed by the manager.

        :return: A list of artifacts.
        """
        return self._artifacts

    @property
    def results(self) -> dict:
        """
        Get the results that were packed by the manager.

        :return: A results dictionary.
        """
        return self._results

    def collect_packagers(
        self, packagers: list[Union[type[Packager], str]], default_priority: int = 5
    ):
        """
        Collect the provided packagers. Packagers passed as module paths are imported and validated to be of type
        `Packager`. If it's needed to import all packagers from a module, use the module path with an asterisk
        "*" at the end. (A packager with a name that starts with an underscore '_' is not collected.)

        Notice: Only packagers that are declared in the module are collected (packagers imported in the module scope
        aren't collected). For example::

            from mlrun import Packager
            from x import XPackager


            class YPackager(Packager):
                pass

        Only "YPackager" is collected since it is declared in the module, but not "XPackager", which is only imported.

        :param packagers:        List of packagers to add.
        :param default_priority: The default priority for the packagers that don't have a set priority (equals to ...).

        :raise MLRunPackageCollectingError: In case the packager could not be collected.
        """
        # Collect the packagers:
        for packager in packagers:
            # If it's a string, it's the module path of the class, so we import it:
            if isinstance(packager, str):
                # TODO: For supporting Hub packagers, if the string is a hub url, then look in the labels for the
                #       packagers to import and import the function as a module.
                # Import the module:
                module_name, class_name = self._split_module_path(module_path=packager)
                try:
                    module = importlib.import_module(module_name)
                except ModuleNotFoundError as module_not_found_error:
                    raise MLRunPackageCollectionError(
                        f"The packager '{class_name}' could not be collected from the module '{module_name}' as it "
                        f"cannot be imported: {module_not_found_error}"
                    ) from module_not_found_error
                # Check if needed to import all packagers from the given module:
                if class_name == "*":
                    # Get all the packagers from the module and collect them (this time they will be sent as `Packager`
                    # types to the method):
                    self.collect_packagers(
                        packagers=[
                            member
                            for _, member in inspect.getmembers(
                                module,
                                lambda m: (
                                    # Validate it is declared in the module:
                                    hasattr(m, "__module__")
                                    and m.__module__ == module.__name__
                                    # Validate it is a `Packager`:
                                    and isinstance(m, type)
                                    and issubclass(m, Packager)
                                    # Validate it is not a "protected" `Packager`:
                                    and not m.__name__.startswith("_")
                                ),
                            )
                        ]
                    )
                    # Collected from the previous call, continue to the next packager in the list:
                    continue
                # Import the packager and continue like as if it was given as a type:
                try:
                    packager = getattr(module, class_name)
                except AttributeError as attribute_error:
                    raise MLRunPackageCollectionError(
                        f"The packager '{class_name}' could not be collected as it does not exist in the module "
                        f"'{module.__name__}': {attribute_error}"
                    ) from attribute_error
            # Validate the class given is a `Packager` type:
            if not issubclass(packager, Packager):
                raise MLRunPackageCollectionError(
                    f"The packager '{packager.__name__}' could not be collected as it is not a `mlrun.Packager`."
                )
            # Initialize the packager class:
            packager = packager()
            # Set default priority in case it is not set in the packager's class:
            if packager.priority is ...:
                packager.priority = default_priority
            # Collect the packager (putting him first in the list for highest priority:
            self._packagers.insert(0, packager)
            # For debugging, we'll print the collected packager:
            logger.debug(
                f"The packagers manager collected the packager: {str(packager)}"
            )

        # Sort the packagers:
        self._packagers.sort()

    def pack(
        self, obj: Any, log_hint: dict[str, str]
    ) -> Union[Artifact, dict, None, list[Union[Artifact, dict, None]]]:
        """
        Pack an object using one of the manager's packagers. A `dict` ("**") or `list` ("*") unpacking syntax in the
        log hint key packs the objects within them in separate packages.

        :param obj:      The object to pack as an artifact.
        :param log_hint: The log hint to use.

        :return: The packaged artifact or result. None is returned if there was a problem while packing the object. If
                 a prefix of dict or list unpacking was provided in the log hint key, a list of all the arbitrary number
                 of packaged objects is returned.

        :raise MLRunInvalidArgumentError: If the key in the log hint instructs to log an arbitrary number of artifacts
                                          but the object type does not match the "*" or "**" used in the key.
        :raise MLRunPackagePackingError:  If there was an error during the packing.
        """
        # Get the key to see if needed to pack arbitrary number of objects via list or dict prefixes:
        log_hint_key = log_hint[LogHintKey.KEY]
        if log_hint_key.startswith("**"):
            # A dictionary unpacking prefix was given, validate the object is a dictionary and prepare the objects to
            # pack with their keys:
            if not isinstance(obj, dict):
                raise MLRunInvalidArgumentError(
                    f"The log hint key '{log_hint_key}' has a dictionary unpacking prefix ('**') to log arbitrary "
                    f"number of objects within the dictionary, but a dictionary was not provided, the given object is "
                    f"of type '{self._get_type_name(type(obj))}'. The object is ignored, to log it, please remove the "
                    f"'**' prefix from the key."
                )
            objects_to_pack = {
                f"{log_hint_key[len('**'):]}{dict_key}": dict_obj
                for dict_key, dict_obj in obj.items()
            }
        elif log_hint_key.startswith("*"):
            # An iterable unpacking prefix was given, validate the object is iterable and prepare the objects to pack
            # with their keys:
            is_iterable = True
            try:
                for _ in obj:
                    break
            except TypeError:
                is_iterable = False
            if not is_iterable:
                raise MLRunInvalidArgumentError(
                    f"The log hint key '{log_hint_key}' has an iterable unpacking prefix ('*') to log arbitrary number "
                    f"of objects within it (like a `list` or `set`), but an iterable object was not provided, the "
                    f"given object is of type '{self._get_type_name(type(obj))}'. The object is ignored, to log it, "
                    f"please remove the '*' prefix from the key."
                )
            objects_to_pack = {
                f"{log_hint_key[len('*'):]}{i}": obj_i for i, obj_i in enumerate(obj)
            }
        else:
            # A single object is required to be packaged:
            objects_to_pack = {log_hint_key: obj}

        # Go over the collected keys and objects and pack them:
        packages = []
        for key, per_key_obj in objects_to_pack.items():
            # Edit the key in the log hint:
            per_key_log_hint = log_hint.copy()
            per_key_log_hint[LogHintKey.KEY] = key
            # Pack and collect the package:
            try:
                packages.append(self._pack(obj=per_key_obj, log_hint=per_key_log_hint))
            except Exception as exception:
                raise MLRunPackagePackingError(
                    f"An exception was raised during the packing of '{per_key_log_hint}': {exception}"
                ) from exception

        # If multiple packages were packed, return a list, otherwise return the single package:
        return packages if len(packages) > 1 else packages[0]

    def unpack(self, data_item: DataItem, type_hint: type) -> Any:
        """
        Unpack an object using one of the manager's packagers. The data item can be unpacked in two ways:

        * As a package: If the data item contains a package and the type hint provided is equal to the object
          type noted in the package. Or, if it's a package and a type hint was not provided.
        * As a data item: If the data item is not a package or the type hint provided is not equal to the one noted in
          the package.

        If the type hint is a `mlrun.DataItem` then it won't be unpacked.

        Notice: It is not recommended to use a different packager than the one that originally packed the object to
        unpack it. A warning displays in that case.

        :param data_item: The data item holding the package.
        :param type_hint: The type hint to parse the data item as.

        :return: The unpacked object parsed as type hinted.
        """
        # Check if `DataItem` is hinted - meaning the user can expect a data item and do not want to unpack it:
        if TypeHintUtils.is_matching(object_type=DataItem, type_hint=type_hint):
            return data_item

        # Set variables to hold the manager notes and packager instructions:
        artifact_key = None
        packaging_instructions = None

        # Try to get the notes and instructions (can be found only in artifacts but data item may be a simple path/url):
        if data_item.get_artifact_type():
            # Get the artifact object in the data item:
            artifact, _ = store_manager.get_store_artifact(url=data_item.artifact_url)
            verify_target_path(artifact)
            # Get the key from the artifact's metadata and instructions from the artifact's spec:
            artifact_key = artifact.metadata.key
            packaging_instructions = artifact.spec.unpackaging_instructions

        # Unpack:
        try:
            if packaging_instructions:
                # The data item is a package and the object type is equal or part of the type hint (part of is in case
                # of a `typing.Union` for example):
                return self._unpack_package(
                    data_item=data_item,
                    artifact_key=artifact_key,
                    packaging_instructions=packaging_instructions,
                    type_hint=type_hint,
                )
            # The data item is not a package or the object type is not equal or part of the type hint:
            return self._unpack_data_item(
                data_item=data_item,
                type_hint=type_hint,
            )
        except Exception as exception:
            raise MLRunPackageUnpackingError(
                f"An exception was raised during the unpacking of '{data_item.key}': {exception}"
            ) from exception

    def link_packages(
        self,
        additional_artifact_uris: dict,
        additional_results: dict,
    ):
        """
        Link packages to each other according to the provided extra data and metrics spec keys. A future link is
        marked with ellipses (...). If no link is found, None is used and a warning is printed.

        :param additional_artifact_uris:    Additional artifact URIs to link (should come from an `mlrun.MLClientCtx`).
        :param additional_results:          Additional results to link (should come from an `mlrun.MLClientCtx`).
        """
        # Join the manager's artifacts and results with the additional ones to look for a link in all of them:
        joined_results = {**additional_results, **self.results}

        # Go over the artifacts and link:
        for artifact in self.artifacts:
            # Go over the extra data keys:
            for key in artifact.spec.extra_data:
                # Future link is marked with ellipses (...):
                if artifact.spec.extra_data[key] is ...:
                    # Look for an artifact or result with this key to link it:
                    extra_data = self._look_for_extra_data(
                        key=key,
                        artifacts=self.artifacts,
                        artifact_uris=additional_artifact_uris,
                        results=joined_results,
                    )
                    # Print a warning if a link is missing:
                    if extra_data is None:
                        logger.warn(
                            f"Could not find {key} to link as extra data for {artifact.key}."
                        )
                    # Link it (None will be used in case it was not found):
                    artifact.spec.extra_data[key] = extra_data
            # Go over the metrics keys if available (`ModelArtifactSpec` has a metrics property that may be waiting for
            # values from logged results):
            if hasattr(artifact.spec, "metrics"):
                for key in artifact.spec.metrics:
                    # Future link is marked with ellipses (...):
                    if artifact.spec.metrics[key] is ...:
                        # Link it (None will be used in case it was not found):
                        artifact.spec.metrics[key] = joined_results.get(key, None)

    def clear_packagers_outputs(self):
        """
        Clear the outputs of all packagers. This method should be called at the end of the run, only after logging all
        artifacts, to ensure that files that require uploading have already been uploaded.
        """
        for packager in self._get_packagers_with_default_packager():
            for path in packager.future_clearing_path_list:
                if not os.path.exists(path):
                    continue
                if os.path.isdir(path):
                    shutil.rmtree(path)
                else:
                    os.remove(path)
            packager.future_clearing_path_list.clear()

    class _InstructionsNotesKey:
        """
        Library of keys for the packager instructions to be added to the packed artifact's spec.
        """

        PACKAGER_NAME = "packager_name"
        OBJECT_TYPE = "object_type"
        ARTIFACT_TYPE = "artifact_type"
        INSTRUCTIONS = "instructions"

    def _get_packagers_with_default_packager(self) -> list[Packager]:
        """
        Get the full list of packagers - the collected packagers and the default packager (located at last place in the
        list - the lowest priority).

        :return: A list of the manager's packagers with the default packager.
        """
        return [*self._packagers, self._default_packager]

    def _get_packager_by_name(self, name: str) -> Union[Packager, None]:
        """
        Look for a packager with the given name and return it.

        If a packager was not found None will be returned.

        :param name: The name of the packager to get.

        :return: The found packager or None if it wasn't found.
        """
        # Look for a packager by exact name:
        for packager in self._get_packagers_with_default_packager():
            if packager.__class__.__name__ == name:
                return packager

        # No packager was found:
        logger.warn(f"The packager '{name}' was not found.")
        return None

    def _get_packager_for_packing(
        self,
        obj: Any,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> Union[Packager, None]:
        """
        Look for a packager that can pack the provided object as the provided artifact type.

        If a packager was not found None will be returned.

        :param obj:            The object to pack.
        :param artifact_type:  The artifact type the packager to get should pack / unpack as.
        :param configurations: The log hint configurations passed by the user.

        :return: The found packager or None if it wasn't found.
        """
        # Look for a packager for the combination of object nad artifact type:
        for packager in self._packagers:
            if packager.is_packable(
                obj=obj, artifact_type=artifact_type, configurations=configurations
            ):
                return packager

        # No packager was found:
        return None

    def _get_packager_for_unpacking(
        self,
        data_item: Any,
        type_hint: type,
        artifact_type: Optional[str] = None,
    ) -> Union[Packager, None]:
        """
        Look for a packager that can unpack the data item of the given type hint as the provided artifact type.

        If a packager was not found None will be returned.

        :param data_item:     The data item to unpack.
        :param type_hint:     The type hint the packager to get should handle.
        :param artifact_type: The artifact type the packager to get should pack / unpack as.

        :return: The found packager or None if it wasn't found.
        """
        # Look for a packager for the combination of object type nad artifact type:
        for packager in self._packagers:
            if packager.is_unpackable(
                data_item=data_item, type_hint=type_hint, artifact_type=artifact_type
            ):
                return packager

        # No packager was found:
        return None

    def _pack(self, obj: Any, log_hint: dict) -> Union[Artifact, dict, None]:
        """
        Pack an object using one of the manager's packagers.

        :param obj:      The object to pack as an artifact.
        :param log_hint: The log hint to use.

        :return: The packaged artifact or result. None is returned if there was a problem while packing the object.
        """
        # Get the artifact type (if user didn't pass any, the packager will use its configured default) and key:
        artifact_type = log_hint.pop(LogHintKey.ARTIFACT_TYPE, None)
        key = log_hint.pop(LogHintKey.KEY, None)

        # Get a packager:
        packager = self._get_packager_for_packing(
            obj=obj, artifact_type=artifact_type, configurations=log_hint
        )
        if packager is None:
            if self._default_packager.is_packable(
                obj=obj, artifact_type=artifact_type, configurations=log_hint
            ):
                logger.info(f"Using the default packager to pack the object '{key}'")
                packager = self._default_packager
            else:
                raise MLRunPackagePackingError(
                    f"No packager was found for the combination of "
                    f"'object_type={self._get_type_name(typ=type(obj))}' and 'artifact_type={artifact_type}'."
                )

        # Use the packager to pack the object:
        packed_object = packager.pack(
            obj=obj, key=key, artifact_type=artifact_type, configurations=log_hint
        )

        # If the packed object is a result, return it as is:
        if isinstance(packed_object, dict):
            # Collect the result and return:
            self._results.update(packed_object)
            return packed_object

        # It is an artifact, continue with the packaging:
        artifact, instructions = packed_object

        # Prepare the manager's unpackaging instructions:
        unpackaging_instructions = {
            self._InstructionsNotesKey.PACKAGER_NAME: packager.__class__.__name__,
            self._InstructionsNotesKey.OBJECT_TYPE: self._get_type_name(typ=type(obj)),
            self._InstructionsNotesKey.ARTIFACT_TYPE: (
                artifact_type
                if artifact_type
                else packager.get_default_packing_artifact_type(obj=obj)
            ),
            self._InstructionsNotesKey.INSTRUCTIONS: instructions,
        }

        # Set the instructions in the artifact's spec:
        artifact.spec.unpackaging_instructions = unpackaging_instructions

        # Collect the artifact and return:
        self._artifacts.append(artifact)
        return artifact

    def _unpack_package(
        self,
        data_item: DataItem,
        artifact_key: str,
        packaging_instructions: dict,
        type_hint: type,
    ) -> Any:
        """
        Unpack a data item as a package using the given notes.

        :param data_item:              The data item to unpack.
        :param artifact_key:           The artifact's key (used only to raise a meaningful error message in case of an
                                       error).
        :param packaging_instructions: The manager's noted instructions.
        :param type_hint:              The user's type hint.

        :return: The unpacked object.

        :raise MLRunPackageUnpackingError: If there is no packager with the given name.
        """
        # Extract the packaging instructions:
        packager_name = packaging_instructions[self._InstructionsNotesKey.PACKAGER_NAME]
        try:
            # For validation, we'll try to get the type of the original packaged object. The original object type might
            # not be available for 3 reasons:
            # 1. The user is trying to parse the data item to a different type than the one it was packaged - meaning it
            #    is ok to be missing, the method will call `unpack_data_item` down the road.
            # 2. The interpreter does not have the required module to unpack this object meaning it will not have the
            #    original packager as well, so it will try to use another package before raising an error.
            # 3. An edge case where the user declared the class at the MLRun function itself. Read the long warning to
            #    understand more.
            self._get_type_from_name(
                type_name=packaging_instructions[self._InstructionsNotesKey.OBJECT_TYPE]
            )
        except ModuleNotFoundError:
            logger.warn(
                f"Could not import the original type "
                f"('{packaging_instructions[self._InstructionsNotesKey.OBJECT_TYPE]}') of the input artifact "
                f"'{artifact_key}' due to a `ModuleNotFoundError`.\n"
                f"Note: If you wish to parse the input to a different type (which is not recommended) you may ignore "
                f"this warning. Otherwise, make sure the interpreter has the required module to import the type.\n"
                f"If it does, you probably implemented the class at the same file of your MLRun function, making "
                f"Python collect it twice: one from the object's own Packager class and another from the function "
                f"code. When MLRun is converting code to a MLRun function, it counts on it to be able to be imported "
                f"as a stand alone file. If other classes (like the packager who imports it) require objects declared "
                f"in this file, it is no longer stand alone. For example:\n\n"
                f""
                f"Let us look at a file '/src/my_module/my_file.py':"
                f"\tclass MyClass:\n"
                f"\t\tpass\n\n"
                f"\tclass MyClassPackager(Packager):\n"
                f"\t\tPACKABLE_OBJECT_TYPE = MyClass\n\n"
                f""
                f"The packager of this class will have the class variable `PACKABLE_OBJECT_TYPE=MyClass` where "
                f"`MyClass`'s module is `src.my_module.my_file.MyClass` because it is being collected from the repo "
                f"downloaded with the project.\n"
                f"But, if creating a MLRun function of '/src/my_module/my_file.py', then 'my_file.py' will be imported "
                f"as a stand alone module, making the same class to be imported twice: one time as `my_file.MyClass` "
                f"from the stand alone function, and another from the packager who has the correct full module path: "
                f"`src.my_module.my_file.MyClass`. This will cause both classes, although the same, to be not equal "
                f"and the first one to be not even importable outside the scope of 'my_file.py' - yielding this "
                f"warning."
            )
        artifact_type = packaging_instructions[self._InstructionsNotesKey.ARTIFACT_TYPE]
        instructions = (
            packaging_instructions[self._InstructionsNotesKey.INSTRUCTIONS] or {}
        )

        # Get the original packager by its name:
        packager = self._get_packager_by_name(name=packager_name)

        # Check if the original packager can be used (the user do not count on parsing to a different type):
        unpack_as_package = False
        if packager is None:
            # The original packager was not found, the user either did not add the custom packager or perhaps wants
            # to unpack the data item as a different type than the original one. We will warn and continue to unpack as
            # a non-package data item:
            logger.warn(
                f"{artifact_key} was originally packaged by a packager of type '{packager_name}' but it "
                f"was not found. Custom packagers should be added to the project running the function "
                f"using the `add_custom_packager` method and make sure the function was set in the project "
                f"with the attribute 'with_repo=True`.\n"
                f"MLRun will try to unpack according to the provided type hint in code."
            )
        elif type_hint is None:
            # User count on the type noted in the package, so we unpack it as is:
            unpack_as_package = True
        else:
            # A type hint is provided, check if the type hint is packable by the packager:
            type_hints = {type_hint}
            while not unpack_as_package and len(type_hints) > 0:
                # Check for each hint (one match is enough):
                for hint in type_hints:
                    if packager.is_unpackable(
                        data_item=data_item, type_hint=hint, artifact_type=artifact_type
                    ):
                        unpack_as_package = True
                        break
                if not unpack_as_package:
                    # Reduce the hints and continue:
                    type_hints = TypeHintUtils.reduce_type_hint(type_hint=type_hints)
            if not unpack_as_package:
                # They are not equal, so we can't count on the original packager noted on the package as the user
                # require different type, so we unpack as data item:
                logger.warn(
                    f"{artifact_key} was originally packaged by '{packager_name}' but the type hint given to "
                    f"unpack it as '{type_hint}' is not supported by it. MLRun will try to look for a matching "
                    f"packager to the type hint instead. Note: it is not recommended to parse an object from type to "
                    f"type using the unpacking mechanism of packagers as unknown behavior might happen."
                )

        # Unpack:
        if unpack_as_package:
            return packager.unpack(
                data_item=data_item,
                artifact_type=artifact_type,
                instructions=instructions,
            )
        return self._unpack_data_item(data_item=data_item, type_hint=type_hint)

    def _unpack_data_item(self, data_item: DataItem, type_hint: type):
        """
        Unpack a data item to the desired hinted type. In case the type hint includes multiple types (as in the case of
        `typing.Union`), the manager goes over the types, and reduces them while looking for the first packager that
        can successfully unpack the data item.

        :param data_item: The data item to unpack.
        :param type_hint: The type hint to unpack it to.

        :return: The unpacked object.

        :raise MLRunPackageUnpackingError: If there is no packager that supports the provided type hint.
        """
        # Prepare a list of a packager and exception string for all the failures in case there was no fitting packager:
        found_packagers: list[tuple[Packager, str]] = []

        # Try to unpack as one of the possible types in the type hint:
        possible_type_hints = {type_hint}
        while len(possible_type_hints) > 0:
            for hint in possible_type_hints:
                # Get the packager by the given type:
                packager = self._get_packager_for_unpacking(
                    data_item=data_item, type_hint=hint
                )
                if packager is None:
                    # No packager was found that supports this hinted type:
                    continue
                # Unpack:
                try:
                    return packager.unpack(
                        data_item=data_item,
                        instructions={},
                    )
                except Exception:
                    # Could not unpack as the reduced type hint, collect the exception and go to the next one:
                    found_packagers.append((packager, traceback.format_exc()))
            # Reduce the type hint list and continue:
            possible_type_hints = TypeHintUtils.reduce_type_hint(
                type_hint=possible_type_hints
            )

        # Check the default packager:
        logger.info(
            f"Trying to use the default packager to unpack the data item '{data_item.key}'"
        )
        try:
            return self._default_packager.unpack(
                data_item=data_item,
                artifact_type=None,
                instructions={},
            )
        except Exception:
            found_packagers.append((self._default_packager, traceback.format_exc()))

        # The method did not return until this point, raise an error:
        raise MLRunPackageUnpackingError(
            f"Could not unpack data item with the hinted type '{type_hint}'. The following packagers were tried to "
            f"be used to unpack it but raised the exceptions joined:\n\n"
            + "\n".join(
                [
                    f"Found packager: '{packager}'\nException: {exception}\n"
                    for packager, exception in found_packagers
                ]
            )
        )

    @staticmethod
    def _look_for_extra_data(
        key: str,
        artifacts: list[Artifact],
        artifact_uris: dict,
        results: dict,
    ) -> Union[Artifact, str, int, float, None]:
        """
        Look for an extra data item (artifact or result) by given key. If not found, None is returned.

        :param key:             Key to look for.
        :param artifacts:       Artifacts to look in.
        :param artifact_uris:   Artifacts URIs to look in.
        :param results:         Results to look in.

        :return: The artifact or result with the same key or None if not found.
        """
        artifact_uris = artifact_uris or {}
        for _key, uri in artifact_uris.items():
            if key == _key:
                try:
                    return get_store_resource(uri)
                except mlrun.errors.MLRunNotFoundError as exc:
                    logger.warn(
                        f"Artifact {key=} not found when looking for extra data",
                        exc=mlrun.errors.err_to_str(exc),
                    )
                    return None

        # Look in the artifacts:
        for artifact in artifacts:
            if key == artifact.key:
                return artifact

        # Look in the results:
        return results.get(key, None)

    @staticmethod
    def _split_module_path(module_path: str) -> tuple[str, str]:
        """
        Split a module path to the module name and the class name. Inner classes are not supported.

        :param module_path: The module path to split.

        :return: A tuple of strings of the module name and the class name.
        """
        # Set the main script module in case there is no module to be found:
        if "." not in module_path:
            module_path = f"__main__.{module_path}"

        # Split and return:
        module_name, class_name = module_path.rsplit(".", 1)
        return module_name, class_name

    @staticmethod
    def _get_type_name(typ: type) -> str:
        """
        Get an object type full name - its module path. For example, the name of a pandas data frame is "DataFrame"
        but its full name (module path) is: "pandas.core.frame.DataFrame".

        Notice: Type hints are not an object type. They are, as their name suggests, only hints. As such, typing hints
        should not be given to this function (they do not have '__name__' and '__qualname__' attributes for example).

        :param typ: The object's type to get its full name.

        :return: The object's type full name.
        """
        # Get the module name:
        module_name = typ.__module__ if hasattr(typ, "__module__") else ""

        # Get the type's (class) name
        class_name = typ.__qualname__ if hasattr(typ, "__qualname__") else typ.__name__

        return f"{module_name}.{class_name}" if module_name else class_name

    @staticmethod
    def _get_type_from_name(type_name: str) -> type:
        """
        Get the type object out of the given module path. The module must be a full module path (for example:
        "pandas.DataFrame" and not "DataFrame") otherwise it assumes to be from the local run module - __main__.

        :param type_name: The type full name (module path) string.

        :return: The represented type as imported from its module.
        """
        module_name, class_name = PackagersManager._split_module_path(
            module_path=type_name
        )
        module = importlib.import_module(module_name)
        return getattr(module, class_name)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from typing import Optional

import mlrun

from .._common import LoggingMode, MLRunInterface
from .._ml_common import MLModelHandler, MLPlan, MLPlanStages, MLProducer
from .estimator import Estimator
from .metric import Metric
from .utils import SKLearnTypes, SKLearnUtils


class SKLearnMLRunInterface(MLRunInterface, ABC):
    """
    Interface for adding MLRun features for machine learning SciKit-Learn API.
    """

    # MLRun's context default name:
    DEFAULT_CONTEXT_NAME = "mlrun-sklearn"

    # Attributes to be inserted so the MLRun interface will be fully enabled.
    _PROPERTIES = {
        # A model handler instance with the model for logging / updating the model (if not provided the model won't be
        # logged / updated at the end of training / testing):
        "_model_handler": None,  # type: MLModelHandler
        # A producer instance for logging this model's training / evaluation artifacts:
        "_producer": None,  # type: MLProducer
        # An estimator instance for logging this model's training / evaluation metrics results:
        "_mlrun_estimator": None,  # type: Estimator
        # The test set (For validation post training or evaluation post prediction):
        "_x_test": None,  # type: SKLearnTypes.DatasetType
        "_y_test": None,  # type: SKLearnTypes.DatasetType
    }
    _METHODS = [
        "configure_logging",
        "_pre_fit",
        "_post_fit",
        "_pre_predict",
        "_post_predict",
    ]

    # Attributes to replace so the MLRun interface will be fully enabled.
    _REPLACED_METHODS = ["fit", "predict", "predict_proba"]

    @classmethod
    def add_interface(
        cls,
        obj: SKLearnTypes.ModelType,
        restoration: SKLearnTypes.MLRunInterfaceRestorationType = None,
    ):
        """
        Enrich the object with this interface properties, methods and functions, so it will have SciKit-Learn MLRun's
        features.

        :param obj:                     The model object to enrich his interface.
        :param restoration: Restoration information tuple as returned from 'remove_interface' in order to
                                        add the interface in a certain state.
        """
        # Check if the given model has the 'predict_proba' method to replace:
        if not hasattr(obj, "predict_proba"):
            cls._REPLACED_METHODS.remove("predict_proba")

        # Add the interface to the model:
        super().add_interface(obj=obj, restoration=restoration)

        # Restore the '_REPLACED_METHODS' list for next models:
        if "predict_proba" not in cls._REPLACED_METHODS:
            cls._REPLACED_METHODS.append("predict_proba")

        # Setup a default producer and estimator:
        if obj._producer is None:
            obj._producer = MLProducer()
            obj._mlrun_estimator = Estimator()

    @classmethod
    def mlrun_fit(cls):
        """
        MLRun's common ML API fit wrapper. It will run a validation post training if provided the required attributes.

        Unsupervised learning algorithms won't be using 'y' in their 'fit' method, but for consistency reasons (as
        written in SciKit-Learn's API documentation): the signature of their 'fit' method will still have 'y'. For that
        we do not need another wrapper that support unsupervised learning models.
        """

        def wrapper(
            self: SKLearnTypes.ModelType,
            X: SKLearnTypes.DatasetType,  # noqa: N803 - should be lowercase "x", kept for BC
            y: SKLearnTypes.DatasetType = None,
            *args,
            **kwargs,
        ):
            # Restore the prediction methods as fit will use them:
            cls._restore_attribute(obj=self, attribute_name="predict")
            if hasattr(self, "predict_proba"):
                cls._restore_attribute(obj=self, attribute_name="predict_proba")

            # Call the pre fit method:
            self._pre_fit(x=X, y=y)

            # Call the original fit method:
            result = self.original_fit(X, y, *args, **kwargs)

            # Call the post fit method:
            self._post_fit(x=X, y=y)

            # Replace the prediction methods again:
            cls._replace_function(obj=self, function_name="predict")
            if hasattr(self, "predict_proba"):
                cls._replace_function(obj=self, function_name="predict_proba")
            return result

        return wrapper

    def mlrun_predict(
        self,
        X: SKLearnTypes.DatasetType,  # noqa: N803 - should be lowercase "x", kept for BC
        *args,
        **kwargs,
    ):
        """
        MLRun's wrapper for the common ML API predict method.
        """
        self._pre_predict(x=X, y=self._y_test)

        y_pred = self.original_predict(X, *args, **kwargs)

        self._post_predict(x=X, y=self._y_test, y_pred=y_pred, is_predict_proba=False)

        return y_pred

    def mlrun_predict_proba(
        self,
        X: SKLearnTypes.DatasetType,  # noqa: N803 - should be lowercase "x", kept for BC
        *args,
        **kwargs,
    ):
        """
        MLRun's wrapper for the common ML API predict_proba method.
        """
        self._pre_predict(x=X, y=self._y_test)

        y_pred = self.original_predict_proba(X, *args, **kwargs)

        self._post_predict(x=X, y=self._y_test, y_pred=y_pred, is_predict_proba=True)

        return y_pred

    def configure_logging(
        self,
        context: mlrun.MLClientCtx = None,
        plans: Optional[list[MLPlan]] = None,
        metrics: Optional[list[Metric]] = None,
        x_test: SKLearnTypes.DatasetType = None,
        y_test: SKLearnTypes.DatasetType = None,
        model_handler: MLModelHandler = None,
    ):
        """
        Configure the MLRun logger for this model using the provided context and artifacts plans, metrics and model
        logging attributes.

        :param context:       A MLRun context to log to. By default, uses `mlrun.get_or_create_ctx`
        :param plans:         A list of plans to produce.
        :param metrics:       A list of metrics to calculate.
        :param x_test:        The testing data for producing and calculating artifacts and metrics post training or post
                              prediction. Without this, training validation will not be performed.
        :param y_test:        The test data for producing and calculating artifacts and metrics post training (calling
                              'fit') or post testing (calling 'predict' / 'predict_proba').
        :param model_handler: A model handler for logging the model as a model artifact post training (post calling
                              'fit') or update the existing model artifact post testing (calling 'predict' /
                              'predict_proba').
        """
        # Store the given model handler:
        self._model_handler = model_handler

        # Get the context:
        if context is None:
            context = (
                mlrun.get_or_create_ctx(name=SKLearnMLRunInterface.DEFAULT_CONTEXT_NAME)
                if model_handler.context is None
                else model_handler.context
            )
        self._producer.set_context(context=context)
        self._mlrun_estimator.set_context(context=context)
        self._model_handler.set_context(context=context)

        # Set the logging attributes:
        self._producer.set_plans(plans=plans)
        self._mlrun_estimator.set_metrics(metrics=metrics)

        # Validate that if the prediction probabilities are required, this model has the 'predict_proba' method:
        if (
            self._producer.is_probabilities_required()
            or self._mlrun_estimator.is_probabilities_required()
        ) and not hasattr(self, "predict_proba"):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Some of the metrics and or artifacts required to be calculated and produced require prediction "
                f"probabilities yet this model: '{type(self)}' do not has the 'predict_proba' method."
            )

        # Store the given dataset:
        self._x_test = x_test
        self._y_test = y_test

    def _pre_fit(self, x: SKLearnTypes.DatasetType, y: SKLearnTypes.DatasetType = None):
        """
        Method for creating the artifacts before the fit method.

        :param x: The input dataset to the fit method ('x_train').
        :param y: The input dataset to the fit method ('y_train').
        """
        self._producer.produce_stage(stage=MLPlanStages.PRE_FIT, model=self, x=x, y=y)

    def _post_fit(
        self, x: SKLearnTypes.DatasetType, y: SKLearnTypes.DatasetType = None
    ):
        """
        Method for creating the artifacts after the fit method. If a validation set is available, the method will start
        a validation process calling predict - creating and calculating validation artifacts and metrics.

        :param x: The input dataset to the fit method ('x_train').
        :param y: The input dataset to the fit method ('y_train').
        """
        # The model is done training, log all artifacts post fit:
        self._producer.produce_stage(stage=MLPlanStages.POST_FIT, model=self, x=x, y=y)

        # If there is a validation set, run validation:
        if self._x_test is not None:
            self._producer.produce_stage(
                stage=MLPlanStages.PRE_PREDICT,
                model=self,
                x=self._x_test,
                y=self._y_test,
            )
            y_pred = self.predict(self._x_test)
            self._post_predict(
                x=self._x_test,
                y=self._y_test,
                y_pred=y_pred,
                is_predict_proba=False,
            )

        # Log the model with the given attributes:
        if self._model_handler is not None:
            # Set the sample set to the training set if None:
            if self._model_handler.sample_set is None:
                sample_set, target_columns = SKLearnUtils.concatenate_x_y(
                    x=x, y=y, target_columns_names=self._model_handler.target_columns
                )
                self._model_handler.set_target_columns(target_columns=target_columns)
                self._model_handler.set_sample_set(sample_set=sample_set)
            # Log the model:
            self._model_handler.log(
                metrics=self._mlrun_estimator.results,
                artifacts=self._producer.artifacts,
            )
            self._model_handler.context.commit(completed=False)

    def _pre_predict(self, x: SKLearnTypes.DatasetType, y: SKLearnTypes.DatasetType):
        """
        Method for creating the artifacts before the predict method.

        :param x: The input dataset to the predict / predict_proba method ('x_test').
        :param y: The input dataset to the predict / predict_proba method ('y_test').
        """
        # This function is only called for evaluation, then set the mode to the producer and estimator:
        self._producer.set_mode(mode=LoggingMode.EVALUATION)
        self._mlrun_estimator.set_mode(mode=LoggingMode.EVALUATION)

        # Produce and log all the artifacts pre prediction:
        self._producer.produce_stage(
            stage=MLPlanStages.PRE_PREDICT, model=self, x=x, y=y
        )

    def _post_predict(
        self,
        x: SKLearnTypes.DatasetType,
        y: SKLearnTypes.DatasetType,
        y_pred: SKLearnTypes.DatasetType,
        is_predict_proba: bool,
    ):
        """
        Method for creating and calculating the artifacts and metrics after the predict method. This method can be
        called after a user call to predict as part of testing or as part of validation after training (calling fit).

        :param x:                The input dataset to the predict / predict_proba method ('x_test' / 'x_validation').
        :param y:                The input dataset to the predict / predict_proba method ('y_test' / 'y_validation').
        :param y_pred:           The prediction outcome.
        :param is_predict_proba: Whether the prediction returned from predict or predict_proba.
        """
        # Produce and log all the artifacts post prediction:
        self._producer.produce_stage(
            stage=MLPlanStages.POST_PREDICT,
            model=self,
            x=x,
            y=y,
            y_pred=y_pred,
            is_probabilities=is_predict_proba,
        )

        # Calculate and log the metrics results:
        self._mlrun_estimator.estimate(
            y_true=y, y_pred=y_pred, is_probabilities=is_predict_proba
        )

        # If some metrics and / or plans require probabilities, run 'predict_proba':
        if not is_predict_proba and (
            self._producer.is_probabilities_required()
            or self._mlrun_estimator.is_probabilities_required()
        ):
            y_pred_proba = self.predict_proba(x)
            self._producer.produce_stage(
                stage=MLPlanStages.POST_PREDICT,
                is_probabilities=True,
                model=self,
                x=x,
                y=y,
                y_pred=y_pred_proba,
            )
            self._mlrun_estimator.estimate(
                y_true=y, y_pred=y_pred_proba, is_probabilities=True
            )

        # If its part of validation post training, return:
        if self._producer.mode == LoggingMode.TRAINING:
            return

        # Update the model with the testing artifacts and results:
        if self._model_handler is not None:
            self._model_handler.update(
                metrics=self._mlrun_estimator.results,
                artifacts=self._producer.artifacts,
            )
            self._model_handler.context.commit(completed=False)



# Copyright 2018 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

from mlrun.common.schemas.notification import Notification
from mlrun.common.schemas.schedule import ScheduleCronTrigger
from mlrun.common.types import StrEnum


class WorkflowSpec(pydantic.v1.BaseModel):
    name: str
    engine: typing.Optional[str] = None
    code: typing.Optional[str] = None
    path: typing.Optional[str] = None
    args: typing.Optional[dict] = None
    handler: typing.Optional[str] = None
    ttl: typing.Optional[int] = None
    args_schema: typing.Optional[list] = None
    schedule: typing.Union[str, ScheduleCronTrigger] = None
    run_local: typing.Optional[bool] = None
    image: typing.Optional[str] = None
    workflow_runner_node_selector: typing.Optional[dict[str, str]] = None


class WorkflowRequest(pydantic.v1.BaseModel):
    spec: typing.Optional[WorkflowSpec] = None
    arguments: typing.Optional[dict] = None
    artifact_path: typing.Optional[str] = None
    source: typing.Optional[str] = None
    run_name: typing.Optional[str] = None
    namespace: typing.Optional[str] = None
    notifications: typing.Optional[list[Notification]] = None


class RerunWorkflowRequest(pydantic.v1.BaseModel):
    run_name: typing.Optional[str] = None
    run_id: typing.Optional[str] = None
    notifications: typing.Optional[list[Notification]] = None
    workflow_runner_node_selector: typing.Optional[dict[str, str]] = None
    original_workflow_runner_uid: typing.Optional[str] = None
    original_workflow_name: typing.Optional[str] = None
    rerun_index: typing.Optional[int] = None


class WorkflowResponse(pydantic.v1.BaseModel):
    project: str = None
    name: str = None
    status: str = None
    run_id: typing.Optional[str] = None
    schedule: typing.Union[str, ScheduleCronTrigger] = None


class GetWorkflowResponse(pydantic.v1.BaseModel):
    workflow_id: str = None


class EngineType(StrEnum):
    LOCAL = "local"
    REMOTE = "remote"
    KFP = "kfp"
    REMOTE_KFP = "remote:kfp"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import numpy as np
import plotly.graph_objects as go

import mlrun
from mlrun.artifacts import Artifact, PlotlyArtifact

from .logger import Logger


class MLRunLogger(Logger):
    """
    MLRun logger is logging the information collected during training of the base logger and logging it to MLRun using
    an MLRun context.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx,
    ):
        """
        Initialize the MLRun logging interface to work with the given context.

        :param context: MLRun context to log to. The context parameters can be logged as static hyperparameters.
        """
        super().__init__()

        # An MLRun context to log to:
        self._context = context

        # Prepare the artifacts dictionary:
        self._artifacts = {}  # type: Dict[str, Artifact]

    def get_artifacts(self) -> dict[str, Artifact]:
        """
        Get the artifacts created by this logger.

        :return: The artifacts dictionary.
        """
        return self._artifacts

    def get_metrics(self) -> dict[str, float]:
        """
        Generate a metrics summary to log along the model.

        :return: The metrics summary.
        """
        return {
            f"{validation_set}_{metric_name}": results[-1]
            for validation_set, metrics in self._results.items()
            for metric_name, results in metrics.items()
        }

    def log_context_parameters(self):
        """
        Log the context given parameters as static hyperparameters. Should be called once as the context parameters do
        not change.
        """
        for parameter_name, parameter_value in self._context.parameters.items():
            # Check if the parameter is a trackable value:
            if isinstance(parameter_value, (str, bool, float, int)):
                self.log_static_hyperparameter(
                    parameter_name=parameter_name, value=parameter_value
                )
            else:
                # See if its string representation length is below the maximum value length:
                string_value = str(parameter_value)
                if (
                    len(string_value) < 30
                ):  # Temporary to no log to long variables into the UI.
                    # TODO: Make the user specify the parameters and take them all by default.
                    self.log_static_hyperparameter(
                        parameter_name=parameter_name, value=parameter_value
                    )

    def log_iteration_to_context(self):
        """
        Log the information of the last iteration and produce the updated artifacts. Each call will log the following
        information:

        * Results table:

          * Static hyperparameters.
          * Dynamic hyperparameters.
          * Metric results.

        * Plot artifacts:

          * A chart for each of the metrics iteration results.
          * A chart for each of the dynamic hyperparameters values.
        """
        # Log the collected hyperparameters:
        for static_parameter, value in self._static_hyperparameters.items():
            self._context.log_result(static_parameter, value)
        for dynamic_parameter, values in self._dynamic_hyperparameters.items():
            # Log as a result to the context (take the most recent result in the training history (-1 index):
            self._context.log_result(dynamic_parameter, values[-1])
            # Create the plotly artifact:
            artifact = self._produce_convergence_plot_artifact(
                name=dynamic_parameter,
                values=values,
            )
            # Log the artifact:
            self._context.log_artifact(artifact)
            # Collect it for later adding it to the model logging as extra data:
            self._artifacts[artifact.metadata.key] = artifact

        # Log the metrics:
        for metric_name, metric_results in {
            f"{validation_set}_{metric_name}": results
            for validation_set, metrics in self._results.items()
            for metric_name, results in metrics.items()
        }.items():
            # Log as a result to the context:
            self._context.log_result(metric_name, metric_results[-1])
            # Create the plotly artifact:
            artifact = self._produce_convergence_plot_artifact(
                name=f"{metric_name}_plot",
                values=metric_results,
            )
            # Log the artifact:
            self._context.log_artifact(artifact)
            # Collect it for later adding it to the model logging as extra data:
            self._artifacts[artifact.metadata.key] = artifact

        # Commit to update the changes, so they will be available in the MLRun UI:
        self._context.commit(completed=False)

    @staticmethod
    def _produce_convergence_plot_artifact(
        name: str, values: list[float]
    ) -> PlotlyArtifact:
        """
        Produce the convergences for the provided metric according.

        :param name:   The name of the metric / hyperparameter.
        :param values: The values per iteration of the metric / hyperparameter.

        :return: Plotly artifact of the convergence plot.
        """
        # Initialize a plotly figure:
        metric_figure = go.Figure()

        # Add titles:
        metric_figure.update_layout(
            title=f"{re.sub('_', ' ', name).capitalize()}",
            xaxis_title="Iterations",
            yaxis_title="Values",
        )

        # Draw:
        metric_figure.add_trace(
            go.Scatter(
                x=list(np.arange(len(values))),
                y=values,
                mode="lines",
            )
        )

        # Creating the artifact:
        return PlotlyArtifact(
            key=name,
            figure=metric_figure,
        )
